@@ -1,3 +1,5 @@
+use crate::number::Number;
+use crate::vm::char::write_escaped_char;
 use ::lazy_static::lazy_static;
 use std::borrow::Borrow;
 use std::collections::HashSet;
@@ -7,13 +9,17 @@ use std::ops::DerefMut;
 #[derive(Debug, Eq, PartialEq, Hash, Clone)]
 pub enum Cell {
     Bool(bool),
+    Char(char),
     Nil,
-    Number(i64),
+    Number(Number),
     Pair(Box<Cell>, Box<Cell>),
+    String(String),
     Symbol(String),
+    Vector(Vec<Cell>),
 
     // Types that exist in VCell, but need Cell representation for
     // printing purposes. These are never created by the lexer/parser.
+    Continuation,
     Closure,
     Macro,
     Lambda,
@@ -26,6 +32,10 @@ impl Cell {
         Cell::Symbol(val.into())
     }
 
+    pub fn new_string(val: &str) -> Cell {
+        Cell::String(val.into())
+    }
+
     pub fn new_list<T: IntoIterator<Item = Cell>>(iter: T) -> Cell {
         Cell::construct_list(iter, None)
     }
@@ -153,6 +163,10 @@ impl Cell {
         self.is_symbol_str("quote")
     }
 
+    pub fn is_define(&self) -> bool {
+        self.is_symbol_str("define")
+    }
+
     pub fn is_lambda(&self) -> bool {
         self.is_symbol_str("lambda")
     }
@@ -173,69 +187,8 @@ impl Cell {
     /// `cell`
     pub fn is_primitive_symbol(&self) -> bool {
         lazy_static! {
-            static ref PRIMITIVE_SYMBOLS: HashSet<&'static str> = HashSet::from([
-                "car",
-                "cdr",
-                "cons",
-                "define",
-                "eq?",
-                "eqv?",
-                "lambda",
-                "quote",
-                "+",
-                "-",
-                "*",
-                "if",
-                "not",
-                "boolean?",
-                "char?",
-                "list?",
-                "null?",
-                "pair?",
-                "port?",
-                "procedure?",
-                "string?",
-                "symbol?",
-                "vector?"
-            ]);
-        }
-        match self {
-            Cell::Symbol(sym) => {
-                PRIMITIVE_SYMBOLS.contains(sym.as_str()) || self.is_syntactic_keyword()
-            }
-            _ => false,
-        }
-    }
-
-    /// Is Syntactic Keyword
-    ///
-    /// Return true if the given cell is a syntactic keyword.
-    ///
-    /// # Arguments
-    /// `cell`
-    pub fn is_syntactic_keyword(&self) -> bool {
-        lazy_static! {
-            static ref PRIMITIVE_SYMBOLS: HashSet<&'static str> = HashSet::from([
-                "begin!",
-                "case",
-                "define",
-                "delay",
-                "do",
-                "else",
-                "lambda",
-                "if",
-                "cond",
-                "and",
-                "or",
-                "let",
-                "let*",
-                "letrec",
-                "quasiquote",
-                "quote",
-                "unquote",
-                "unquote-splicnig",
-                "set!",
-            ]);
+            static ref PRIMITIVE_SYMBOLS: HashSet<&'static str> =
+                HashSet::from(["define", "lambda", "if", "quote", "set!"]);
         }
         match self {
             Cell::Symbol(sym) => PRIMITIVE_SYMBOLS.contains(sym.as_str()),
@@ -271,9 +224,9 @@ impl Cell {
         }
     }
 
-    pub fn as_number(&self) -> Option<i64> {
+    pub fn as_number(&self) -> Option<Number> {
         match self {
-            Cell::Number(val) => Some(*val),
+            Cell::Number(val) => Some(val.clone()),
             _ => None,
         }
     }
@@ -307,7 +260,13 @@ impl From<&str> for Cell {
 
 impl From<i64> for Cell {
     fn from(val: i64) -> Self {
-        Cell::Number(val)
+        Cell::Number(Number::Fixnum(val))
+    }
+}
+
+impl From<char> for Cell {
+    fn from(val: char) -> Self {
+        Cell::Char(val)
     }
 }
 
@@ -321,6 +280,8 @@ pub struct IntoIter<'a> {
     next: &'a Cell,
 }
 
+impl<'a> ExactSizeIterator for IntoIter<'a> {}
+
 impl<'a> Iterator for IntoIter<'a> {
     type Item = &'a Cell;
 
@@ -339,6 +300,11 @@ impl<'a> Iterator for IntoIter<'a> {
             }
         }
     }
+
+    fn size_hint(&self) -> (usize, Option<usize>) {
+        let len = self.next.len();
+        (len, Some(len))
+    }
 }
 
 impl<'a> IntoIterator for &'a Cell {
@@ -354,6 +320,8 @@ pub struct Iter {
     next: Option<Cell>,
 }
 
+impl ExactSizeIterator for Iter {}
+
 impl Iterator for Iter {
     type Item = Cell;
 
@@ -370,6 +338,14 @@ impl Iterator for Iter {
             }
         }
     }
+
+    fn size_hint(&self) -> (usize, Option<usize>) {
+        let len = match &self.next {
+            Some(cell) => cell.len(),
+            None => 0,
+        };
+        (len, Some(len))
+    }
 }
 
 impl IntoIterator for Cell {
@@ -383,14 +359,10 @@ impl IntoIterator for Cell {
 
 impl Display for Cell {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
-        lazy_static! {
-            static ref QUOTE: Cell = Cell::Symbol("quote".into());
-        }
-
         match self {
             Cell::Pair(car, cdr) => {
                 // sugar quote any list in the exact form (quote x)
-                if **car == *QUOTE && (*cdr).is_pair() && (*cdr).cdr().unwrap().is_nil() {
+                if car.is_quote() && (*cdr).is_pair() && (*cdr).cdr().unwrap().is_nil() {
                     write!(f, "'")?;
                     return std::fmt::Display::fmt(cdr.car().unwrap(), f);
                 }
@@ -400,16 +372,16 @@ impl Display for Cell {
                 loop {
                     match (*cdr).as_ref() {
                         Cell::Nil => {
-                            write!(f, "{})", car)?;
+                            write!(f, "{:#})", car)?;
                             return Ok(());
                         }
                         Cell::Pair(ncar, ncdr) => {
-                            write!(f, "{} ", car)?;
+                            write!(f, "{:#} ", car)?;
                             car = ncar;
                             cdr = ncdr;
                         }
                         _ => {
-                            write!(f, "{} . {})", car, cdr)?;
+                            write!(f, "{:#} . {:#})", car, cdr)?;
                             return Ok(());
                         }
                     }
@@ -418,15 +390,52 @@ impl Display for Cell {
             Cell::Bool(val) => {
                 write!(f, "{}", if *val { "#t" } else { "#f" })
             }
+            Cell::Char(c) => match f.alternate() {
+                false => write!(f, "{}", c),
+                true => write_escaped_char(*c, f),
+            },
             Cell::Number(val) => {
                 write!(f, "{}", val)
             }
+            Cell::String(val) => match f.alternate() {
+                false => write!(f, "{}", val),
+                true => {
+                    write!(f, "\"")?;
+                    for it in val.chars() {
+                        match it {
+                            '"' | '\\' => write!(f, "\\{}", it)?,
+                            '\t' => write!(f, "\\t")?,
+                            '\n' => write!(f, "\\n")?,
+                            '\r' => write!(f, "\\r")?,
+                            _ if it as u32 == 0x7 => write!(f, "\\a")?,
+                            _ if it as u32 == 0x8 => write!(f, "\\b")?,
+                            _ if it.is_ascii_control() => write!(f, "\\x{:x};", it as u32)?,
+                            it => write!(f, "{}", it)?,
+                        };
+                    }
+                    write!(f, "\"")
+                }
+            },
             Cell::Symbol(val) => {
                 write!(f, "{}", val)
             }
             Cell::Nil => {
                 write!(f, "()")
             }
+            Cell::Vector(vector) => {
+                write!(f, "#(")?;
+                for (idx, cell) in vector.iter().enumerate() {
+                    if idx == vector.len() - 1 {
+                        write!(f, "{:#}", cell)?;
+                    } else {
+                        write!(f, "{:#} ", cell)?;
+                    }
+                }
+                write!(f, ")")
+            }
+            Cell::Continuation => {
+                write!(f, "#<continuation>")
+            }
             Cell::Closure => {
                 write!(f, "#<procedure>")
             }
@@ -492,13 +501,27 @@ macro_rules! list {
     }};
 }
 
+#[macro_export]
+macro_rules! vector {
+    () => {
+        Cell::Vector(vec![])
+    };
+    ($($elt:expr),+) => {{
+        let v = vec![$(Cell::from($elt),)+];
+        Cell::Vector(v)
+    }};
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
 
     #[test]
     fn eq() {
-        assert_eq!(Cell::Number(16), Cell::Number(16));
+        assert_eq!(
+            Cell::Number(Number::Fixnum(16)),
+            Cell::Number(Number::Fixnum(16))
+        );
         assert_eq!(Cell::new_symbol("foo"), Cell::new_symbol("foo"));
         assert_eq!(
             Cell::new_list(vec!(Cell::new_symbol("foo"), Cell::new_symbol("bar"))),
@@ -511,27 +534,46 @@ mod tests {
     fn cell_macro() {
         assert_eq!(cell![], Cell::Nil);
         assert_eq!(cell!["foo"], Cell::Symbol("foo".into()));
-        assert_eq!(cell![42], Cell::Number(42));
-        assert_eq!(cell![-42], Cell::Number(-42));
+        assert_eq!(cell![42], Cell::Number(Number::Fixnum(42)));
+        assert_eq!(cell![-42], Cell::Number(Number::Fixnum(-42)));
         assert_eq!(
             cell![0, 1, 2],
-            Cell::new_list(vec!(Cell::Number(0), Cell::Number(1), Cell::Number(2)))
+            Cell::new_list(vec!(
+                Cell::Number(Number::Fixnum(0)),
+                Cell::Number(Number::Fixnum(1)),
+                Cell::Number(Number::Fixnum(2))
+            ))
         );
         assert_eq!(
             cell!["foo", 42],
-            Cell::new_list(vec!(Cell::new_symbol("foo"), Cell::Number(42)))
+            Cell::new_list(vec!(
+                Cell::new_symbol("foo"),
+                Cell::Number(Number::Fixnum(42))
+            ))
         );
         assert_eq!(
             cell!["foo", cell![0, 1, 2]],
             Cell::new_list(vec!(
                 Cell::new_symbol("foo"),
-                Cell::new_list(vec!(Cell::Number(0), Cell::Number(1), Cell::Number(2)))
+                Cell::new_list(vec!(
+                    Cell::Number(Number::Fixnum(0)),
+                    Cell::Number(Number::Fixnum(1)),
+                    Cell::Number(Number::Fixnum(2))
+                ))
             ))
         );
         assert_eq!(list![], Cell::new_list(vec!()));
         assert_eq!(list!["foo"], Cell::new_list(vec!(Cell::new_symbol("foo"))));
     }
 
+    #[test]
+    fn vector_macr() {
+        assert_eq!(
+            vector![1, 2, 3],
+            Cell::Vector(vec![cell![1], cell![2], cell![3]])
+        );
+    }
+
     #[test]
     fn proper_list() {
         assert!(list![1, 2, 3].is_list());
