@@ -7,15 +7,19 @@ use std::str::CharIndices;
 /// by the scanner.
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub enum TokenType {
+    Char,
     Dot,
     False,
     LeftParen,
     Number,
+    NumberPrefix,
     RightParen,
     SingleQuote,
+    String,
     Symbol,
     True,
     WhiteSpace,
+    HashParen,
 }
 
 /// Token
@@ -52,6 +56,18 @@ impl Token {
     pub fn span<'a, 'b>(&'a self, text: &'b str) -> &'b str {
         &text[self.span.0..self.span.1]
     }
+
+    /// span prefix
+    ///
+    /// Given the originally scanned &str, extract any text before this
+    /// token.
+    pub fn span_prefix<'a, 'b>(&'a self, text: &'b str) -> &'b str {
+        &text[0..self.span.0]
+    }
+
+    pub fn is_symbol(&self) -> bool {
+        self.token_type == TokenType::Symbol
+    }
 }
 
 /// Error Type
@@ -59,10 +75,12 @@ impl Token {
 /// The type of error encountered by the scanner.
 #[derive(thiserror::Error, Debug, Eq, PartialEq)]
 pub enum Error {
-    #[error("vectors are not supported")]
-    VectorsNotSupported,
+    #[error("unexpected EOF")]
+    Eof,
     #[error("unexpected character '{0}'")]
     UnexpectedToken(char),
+    #[error("unexpected character following '{0}': '{1}'")]
+    UnexpectedCharacterFollowing(String, String),
 }
 
 /// Scan
@@ -96,6 +114,7 @@ pub fn scan(text: &str) -> Result<Vec<Token>, Error> {
             '(' | ')' | '[' | ']' | '{' | '}' | '\'' => scan_simple_token(&mut cur)?,
             '#' => scan_hash_token(&mut cur)?,
             '.' => scan_dot(&mut cur)?,
+            '"' => scan_string(&mut cur)?,
             _ if is_initial_identifier(c) => scan_symbol(&mut cur)?,
             _ if is_initial_number(c) => scan_number(&mut cur)?,
             _ if c.is_whitespace() => {
@@ -109,21 +128,43 @@ pub fn scan(text: &str) -> Result<Vec<Token>, Error> {
     Ok(tokens)
 }
 
+/// Scan Dot
+///
+/// Scan dot follows these rules:
+///
+/// * If a dot is not immediately followed by a subsequent number or
+///   identifier character, then it is a lone dot.
+/// * If a dot is followed by a subsequent number then it is parsed as
+///   a number. This allows numbers such as .333 without the leading 0.
+/// * If a number has any subsequent ., then it is downgraded to a symbol.
+/// * Anything else is a symbol.
 fn scan_dot(cur: &mut Peekable<CharIndices>) -> Result<Token, Error> {
-    let start = cur.peek().unwrap().0;
-    let mut end = start;
+    let start = cur.next().unwrap().0;
+    let mut end = start + '.'.len_utf8();
+
+    let mut token_type = match cur.peek() {
+        Some(c) if is_subsequent_number(c.1) => TokenType::Number,
+        Some(c) if is_subsequent_identifier(c.1) => TokenType::Symbol,
+        _ => TokenType::Dot,
+    };
+
     while let Some(&(offset, c)) = cur.peek() {
-        if !is_subsequent_identifier(c) && start != end {
+        if c == '.' {
+            token_type = TokenType::Symbol;
+        }
+        let check = match token_type {
+            TokenType::Symbol => is_subsequent_identifier(c),
+            TokenType::Number => is_subsequent_number(c),
+            _ => false,
+        };
+        if !check && start != end {
             break;
         }
         end = offset + c.len_utf8();
         cur.next();
     }
-    if end - start == 1 {
-        Ok(Token::new((start, end), TokenType::Dot))
-    } else {
-        Ok(Token::new((start, end), TokenType::Symbol))
-    }
+
+    Ok(Token::new((start, end), token_type))
 }
 
 fn scan_simple_token(cur: &mut Peekable<CharIndices>) -> Result<Token, Error> {
@@ -143,13 +184,19 @@ fn scan_simple_token(cur: &mut Peekable<CharIndices>) -> Result<Token, Error> {
 
 fn scan_hash_token(cur: &mut Peekable<CharIndices>) -> Result<Token, Error> {
     let (start, _) = cur.next().unwrap();
-    let (_, c) = cur.next().ok_or(Error::UnexpectedToken('#'))?;
+    let (_, c) = cur
+        .next()
+        .ok_or_else(|| Error::UnexpectedCharacterFollowing('#'.into(), "\\n".into()))?;
 
     match c {
         't' => Ok(Token::new((start, start + 2), TokenType::True)),
         'f' => Ok(Token::new((start, start + 2), TokenType::False)),
-        '(' => Err(Error::VectorsNotSupported),
-        _ => Err(Error::UnexpectedToken('#')),
+        '(' => Ok(Token::new((start, start + 2), TokenType::HashParen)),
+        'e' | 'i' | 'b' | 'o' | 'd' | 'x' => {
+            Ok(Token::new((start, start + 2), TokenType::NumberPrefix))
+        }
+        '\\' => scan_char(cur, start),
+        c => Err(Error::UnexpectedCharacterFollowing('#'.into(), c.into())),
     }
 }
 
@@ -166,6 +213,69 @@ fn scan_symbol(cur: &mut Peekable<CharIndices>) -> Result<Token, Error> {
     Ok(Token::new((start, end), TokenType::Symbol))
 }
 
+fn scan_string(cur: &mut Peekable<CharIndices>) -> Result<Token, Error> {
+    let start = cur.peek().ok_or(Error::Eof)?.0;
+    cur.next();
+
+    let mut escape_next = false;
+    let mut terminated = false;
+    while let Some(&(_, c)) = cur.peek() {
+        let escaping = escape_next;
+        if c == '"' && !escaping {
+            terminated = true;
+            break;
+        }
+        if c == '\\' && !escaping {
+            escape_next = true;
+        } else {
+            escape_next = false;
+        }
+        cur.next();
+    }
+    // empty string
+    if !terminated {
+        return Err(Error::Eof);
+    }
+    let end = cur.next().unwrap().0 + '"'.len_utf8();
+    Ok(Token::new((start, end), TokenType::String))
+}
+
+/// Scan Char
+///
+/// Scan char is called when the scanner has encountered the char prefix
+/// #\. It will have called scan_char after incrementing cursor past the
+/// character prefix, both start and end signifying the start and end of
+/// the prefix token, where end = start+2.
+///
+/// In the example #\space, cur.next() should be 's' and start points to
+/// the the very first character of the prefix '#'.
+///
+/// # Arguments
+/// `cur` - The cursor, pointing at the first character of the
+///   actual character.
+/// `start` - The start of the token.
+fn scan_char(cur: &mut Peekable<CharIndices>, start: usize) -> Result<Token, Error> {
+    let mut end;
+    if let Some((offset, c)) = cur.next() {
+        end = offset + c.len_utf8();
+        if !c.is_ascii_alphabetic() {
+            return Ok(Token::new((start, end), TokenType::Char));
+        }
+    } else {
+        return Err(Error::Eof);
+    }
+
+    while let Some(&(offset, c)) = cur.peek() {
+        if !c.is_ascii_alphanumeric() {
+            break;
+        }
+        end = offset + c.len_utf8();
+        cur.next();
+    }
+
+    Ok(Token::new((start, end), TokenType::Char))
+}
+
 fn scan_number(cur: &mut Peekable<CharIndices>) -> Result<Token, Error> {
     let start = cur.peek().unwrap().0;
     let mut end = start;
@@ -179,22 +289,24 @@ fn scan_number(cur: &mut Peekable<CharIndices>) -> Result<Token, Error> {
     Ok(Token::new((start, end), TokenType::Number))
 }
 
-fn is_initial_number(c: char) -> bool {
+pub fn is_initial_number(c: char) -> bool {
     c.is_digit(10) || c == '+' || c == '-'
 }
 
-fn is_subsequent_number(c: char) -> bool {
-    c.is_digit(10) || c == '.'
+pub fn is_subsequent_number(c: char) -> bool {
+    c.is_digit(10) || c.is_digit(16) || c == '.' || c == '/'
 }
 
-fn is_initial_identifier(c: char) -> bool {
+pub fn is_initial_identifier(c: char) -> bool {
     c.is_alphabetic()
+        || c as u32 > 0xFF
         || c == '!'
         || c == '$'
         || c == '%'
         || c == '&'
         || c == '*'
         || c == '/'
+        || c == '\\'
         || c == ':'
         || c == '<'
         || c == '='
@@ -205,11 +317,11 @@ fn is_initial_identifier(c: char) -> bool {
         || c == '~'
 }
 
-fn is_special_subsequent(c: char) -> bool {
-    c == '+' || c == '-' || c == '.' || c == '@'
+pub fn is_special_subsequent(c: char) -> bool {
+    c == '+' || c == '-' || c == '.' || c == '@' || c == ';'
 }
 
-fn is_subsequent_identifier(c: char) -> bool {
+pub fn is_subsequent_identifier(c: char) -> bool {
     is_initial_identifier(c) || c.is_digit(10) || is_special_subsequent(c)
 }
 
@@ -250,7 +362,8 @@ mod tests {
     fn parens() {
         lexes! {
             "(" => TokenType::LeftParen,
-            ")" => TokenType::RightParen
+            ")" => TokenType::RightParen,
+            "#(" => TokenType::HashParen
         };
     }
 
@@ -261,6 +374,49 @@ mod tests {
         };
     }
 
+    #[test]
+    fn chars() {
+        lexes! {
+            "#\\a" => TokenType::Char,
+            "#\\space" => TokenType::Char,
+            "#\\newline" => TokenType::Char
+        };
+    }
+
+    #[test]
+    fn strings() {
+        assert_eq!(
+            expand(scan(r#""""#).unwrap(), r#""""#),
+            [(r#""""#, TokenType::String)]
+        );
+
+        assert_eq!(
+            expand(scan(r#""a""#).unwrap(), r#""a""#),
+            [(r#""a""#, TokenType::String)]
+        );
+
+        assert_eq!(
+            expand(scan(r#""foo bar baz""#).unwrap(), r#""foo bar baz""#),
+            [("\"foo bar baz\"", TokenType::String)]
+        );
+
+        assert_eq!(
+            expand(
+                scan(r#""The word \"recursion\" has many meanings.""#).unwrap(),
+                r#""The word \"recursion\" has many meanings.""#
+            ),
+            [(
+                r#""The word \"recursion\" has many meanings.""#,
+                TokenType::String
+            )]
+        );
+
+        assert_eq!(
+            expand(scan(r#""\\\\""#).unwrap(), r#""\\\\""#),
+            [((r#""\\\\""#), TokenType::String)]
+        );
+    }
+
     #[test]
     fn numbers() {
         lexes! {
@@ -289,6 +445,9 @@ mod tests {
             ".." => TokenType::Symbol,
             "..." => TokenType::Symbol
         };
+        lexes! {
+            ".100" => TokenType::Number
+        };
     }
 
     #[test]
@@ -317,8 +476,8 @@ mod tests {
         };
 
         fails! {
-            "#(vector)" => Error::VectorsNotSupported,
-            "#b" => Error::UnexpectedToken('#')
+            "#" => Error::UnexpectedCharacterFollowing('#'.into(), "\\n".into()),
+            "#p" => Error::UnexpectedCharacterFollowing('#'.into(), "p".into())
         };
     }
 }
