@@ -1,12 +1,13 @@
 use crate::cell::Cell;
-use crate::vm::environment::{free_symbols, BindingLocation};
+use crate::vm::environment::{free_symbols, internally_defined_symbols, BindingLocation};
 use crate::vm::lambda::Lambda;
 use crate::vm::opcode::OpCode;
 use crate::vm::transform::Transform;
 use crate::vm::vcell::VCell;
 use crate::vm::vcell::VCell::{BasePointerOffset, LexicalEnvSlot};
 use crate::vm::Error::{
-    InvalidArgs, InvalidNumArgs, InvalidSyntactic, LambdaMissingExpression, UnquotedNil,
+    InvalidArgs, InvalidNumArgs, InvalidSyntax, InvalidUsePrimitive, LambdaMissingExpression,
+    UnquotedNil,
 };
 use crate::vm::{Error, Vm};
 use log::trace;
@@ -38,14 +39,15 @@ impl Vm {
     /// `expr` - The expression to compile.
     pub fn compile(&mut self, expr: &Cell) -> Result<Lambda, Error> {
         let mut entry_lambda = Lambda::new(vec![]);
-        let mut lambda = Lambda::new_from_iof(vec![], &entry_lambda, &[], false);
+        let mut lambda = Lambda::new_from_iof(vec![], vec![], &entry_lambda, &[], false);
+        lambda.set_top_level();
         lambda.emit(OpCode::Enter);
         self.compile_expression(&mut lambda, true, expr)?;
         lambda.emit(OpCode::Ret);
         trace!("main: \n{}", self.decompile_text(&lambda));
         let lambda = self.heap.put(lambda);
         entry_lambda.emit(OpCode::PushImmediate);
-        entry_lambda.emit(VCell::FixedNum(0));
+        entry_lambda.emit(VCell::ArgumentCount(0));
         entry_lambda.emit(OpCode::MovImmediate);
         entry_lambda.emit(lambda);
         entry_lambda.emit(VCell::Acc);
@@ -73,11 +75,15 @@ impl Vm {
             Cell::Pair(_, _) => self.compile_procedure_application(lambda, tail, expr),
             Cell::Symbol(_) => self.compile_symbol_expression(lambda, expr),
             Cell::Nil => Err(UnquotedNil),
-            Cell::Number(_)
-            | Cell::Bool(_)
+            Cell::Bool(_)
+            | Cell::Continuation
+            | Cell::Char(_)
+            | Cell::Number(_)
             | Cell::Closure
             | Cell::Macro
+            | Cell::String(_)
             | Cell::Undefined
+            | Cell::Vector(_)
             | Cell::Void
             | Cell::Lambda => self.compile_quote(lambda, expr),
         }
@@ -119,12 +125,6 @@ impl Vm {
             }
         }
 
-        //
-        // if car.deref().is_symbol() {
-        //     let sym_ref = self.heap.put(car);
-        //     self.globenv.get_binding()
-        // }
-
         match proc.deref() {
             Cell::Symbol(proc) => match proc.as_str() {
                 "define" => self.compile_define(lambda, expr),
@@ -132,6 +132,7 @@ impl Vm {
                 "lambda" | "λ" => self.compile_lambda(lambda, expr, false),
                 "quote" => self.compile_quote(lambda, car!(rest)),
                 "if" => self.compile_if(lambda, tail, expr),
+                "set!" => self.compile_set(lambda, tail, expr),
                 _ => self.compile_runtime_procedure_application(lambda, tail, expr),
             },
             _ => self.compile_runtime_procedure_application(lambda, tail, expr),
@@ -150,8 +151,8 @@ impl Vm {
         lambda: &mut Lambda,
         sym: &Cell,
     ) -> Result<(), Error> {
-        if sym.is_syntactic_keyword() {
-            return Err(InvalidSyntactic(sym.to_string()));
+        if sym.is_primitive_symbol() {
+            return Err(InvalidUsePrimitive(sym.to_string()));
         }
         let sym_ref = self.heap.put_cell(sym);
         match lambda.binding_location(&sym_ref) {
@@ -218,21 +219,92 @@ impl Vm {
             }
         };
 
-        if symbol.is_syntactic_keyword() {
-            return Err(InvalidSyntactic(symbol.to_string()));
+        if symbol.is_primitive_symbol() {
+            return Err(InvalidUsePrimitive(symbol.to_string()));
         }
 
-        let sym_ref = self.heap.put_cell(symbol).as_ptr()?;
-        let env_slot = VCell::env_slot(self.globenv.get_binding(sym_ref));
+        let sym_ref = self.heap.put_cell(symbol);
         lambda.emit(OpCode::Mov);
         lambda.emit(VCell::Acc);
-        lambda.emit(env_slot);
+        match lambda.binding_location(&sym_ref) {
+            BindingLocation::Global => {
+                let sym_ref = sym_ref.as_ptr().expect("expected ptr");
+                let env_slot = VCell::env_slot(self.globenv.get_binding(sym_ref));
+                lambda.emit(env_slot);
+            }
+            BindingLocation::Argument(n) => {
+                let arg_offset = 0_i64 - lambda.argc() as i64 + n as i64 + 1;
+                lambda.emit(BasePointerOffset(arg_offset));
+            }
+            BindingLocation::Environment(n) => {
+                lambda.emit(LexicalEnvSlot(n));
+            }
+        }
+
         lambda.emit(OpCode::MovImmediate);
         lambda.emit(VCell::void());
         lambda.emit(VCell::Acc);
         Ok(())
     }
 
+    /// Set
+    ///
+    /// Set is a primitive that provides support for the set! procedure.
+    ///
+    /// A set! must be in the following form:
+    ///
+    /// * (set ⟨variable⟩ ⟨expression⟩)
+    ///
+    /// `lambda` - The lambda to emit bytecode to
+    /// `expr` - (set! variable expression)
+    /// `tail` - whether or not this set is in a tail context
+    pub fn compile_set(
+        &mut self,
+        lambda: &mut Lambda,
+        _tail: bool,
+        expr: &Cell,
+    ) -> Result<(), Error> {
+        let rest = cdr!(expr);
+        let (variable, expression) = match rest.collect_vec().as_slice() {
+            [variable, expression] => (*variable, *expression),
+            _ => {
+                return Err(InvalidNumArgs("set!".into()));
+            }
+        };
+
+        if !variable.is_symbol() || variable.is_primitive_symbol() {
+            return Err(InvalidSyntax(format!(
+                "expected variable, but got {}",
+                variable
+            )));
+        }
+
+        self.compile_expression(lambda, false, expression)?;
+
+        let sym_ref = self.heap.put_cell(variable);
+        lambda.emit(OpCode::Mov);
+        lambda.emit(VCell::Acc);
+        match lambda.binding_location(&sym_ref) {
+            BindingLocation::Global => {
+                let sym_ref = sym_ref.as_ptr().expect("expected ptr");
+                let env_slot = VCell::env_slot(self.globenv.get_binding(sym_ref));
+                lambda.emit(env_slot);
+            }
+            BindingLocation::Argument(n) => {
+                let arg_offset = 0_i64 - lambda.argc() as i64 + n as i64 + 1;
+                lambda.emit(BasePointerOffset(arg_offset));
+            }
+            BindingLocation::Environment(n) => {
+                lambda.emit(LexicalEnvSlot(n));
+            }
+        }
+
+        lambda.emit(OpCode::MovImmediate);
+        lambda.emit(self.heap.put(VCell::Void));
+        lambda.emit(VCell::Acc);
+        Ok(())
+    }
+
     /// Compile Define Syntax
     ///
     /// Compile a top-level macro to a VCell::Transform, storing it in the
@@ -301,10 +373,22 @@ impl Vm {
         let (formal_args, is_vararg) = self.compile_formal_arguments(formal_args)?;
         let free_symbols = free_symbols(expr)?
             .iter()
+            .inspect(|it| trace!("free: {}", it))
+            .map(|sym| self.heap.put_cell(sym))
+            .collect::<Vec<VCell>>();
+        let internally_defined = internally_defined_symbols(body)?
+            .iter()
+            .inspect(|it| trace!("internal: {}", it))
             .map(|sym| self.heap.put_cell(sym))
             .collect::<Vec<VCell>>();
 
-        let mut lambda = Lambda::new_from_iof(formal_args, iof, &free_symbols, is_vararg);
+        let mut lambda = Lambda::new_from_iof(
+            formal_args,
+            internally_defined,
+            iof,
+            &free_symbols,
+            is_vararg,
+        );
         if lambda.is_vararg {
             lambda.emit(OpCode::VarArg);
         }
@@ -359,16 +443,16 @@ impl Vm {
                     symbol.to_string(),
                 ));
             }
-            if symbol.is_syntactic_keyword() {
-                return Err(InvalidSyntactic(symbol.to_string()));
+            if symbol.is_primitive_symbol() {
+                return Err(InvalidUsePrimitive(symbol.to_string()));
             }
             symbols.push(self.heap.put_cell(symbol));
             rest = cdr!(rest);
         }
 
         if rest.is_symbol() {
-            if rest.is_syntactic_keyword() {
-                return Err(InvalidSyntactic(rest.to_string()));
+            if rest.is_primitive_symbol() {
+                return Err(InvalidUsePrimitive(rest.to_string()));
             }
             symbols.push(self.heap.put_cell(rest));
             Ok((symbols, true))
@@ -408,7 +492,7 @@ impl Vm {
         let proc = car!(expr);
         let mut rest = cdr!(expr);
         // Evaluate and push each argument left-to-right
-        let mut n: i64 = 0;
+        let mut n = 0;
         while rest.is_pair() {
             self.compile_expression(lambda, false, rest.car().unwrap())?;
             lambda.emit(OpCode::PushAcc);
@@ -418,7 +502,7 @@ impl Vm {
 
         // Push the argument count
         lambda.emit(OpCode::PushImmediate);
-        lambda.emit(VCell::FixedNum(n));
+        lambda.emit(VCell::ArgumentCount(n));
 
         // Evaluate the procedure to call, and emit a CALL instruction
         self.compile_expression(lambda, false, proc)?;
