@@ -2,6 +2,7 @@ use crate::cell::Cell;
 use crate::vm::lambda::Lambda;
 use crate::vm::vcell::VCell;
 use crate::vm::Error;
+use crate::vm::Error::InvalidDefineSyntax;
 use std::cell::RefCell;
 use std::collections::{HashMap, HashSet};
 
@@ -39,6 +40,12 @@ pub enum BindingSource {
     /// The binding is inherited from the Nth slot of the IOF's
     /// environment
     IofEnvironment(usize),
+
+    /// The binding is an internal definition (i.e. define in the body)
+    ///
+    /// For example, in `(lambda (x) (define y 10) (+ x y))`, y is an
+    /// internal definition.
+    InternalDefinition,
 }
 
 /// Binding Location
@@ -84,20 +91,37 @@ impl EnvironmentMap {
     /// New From IOF
     ///
     /// Create a new lambda, and populate its environment map
-    /// given the IOF and set oxf free symbols.
-    pub fn new_from_iof(_args: &[VCell], iof: &Lambda, free_symbols: &[VCell]) -> EnvironmentMap {
-        let map = free_symbols
+    /// given the IOF and set of free symbols.
+    pub fn new_from_iof(
+        args: &[VCell],
+        internally_defined: &[VCell],
+        iof: &Lambda,
+        free_symbols: &[VCell],
+    ) -> EnvironmentMap {
+        let mut map = args
+            .iter()
+            .enumerate()
+            .map(|it| (it.1.clone(), BindingSource::Argument(it.0)))
+            .collect::<Vec<(VCell, BindingSource)>>();
+
+        internally_defined
+            .iter()
+            .map(|it| (it.clone(), BindingSource::InternalDefinition))
+            .for_each(|binding| map.push(binding));
+
+        free_symbols
             .iter()
             .filter_map(|sym| {
-                if let Some((n, _)) = iof.args.iter().enumerate().find(|(_, it)| *it == sym) {
+                if let Some(slot) = iof.envmap.get_slot(sym) {
+                    Some((sym.clone(), BindingSource::IofEnvironment(slot)))
+                } else if let Some((n, _)) = iof.args.iter().enumerate().find(|(_, it)| *it == sym)
+                {
                     Some((sym.clone(), BindingSource::IofArgument(n)))
                 } else {
-                    iof.envmap
-                        .get_slot(sym)
-                        .map(|n| (sym.clone(), BindingSource::IofEnvironment(n)))
+                    None
                 }
             })
-            .collect::<Vec<(VCell, BindingSource)>>();
+            .for_each(|binding| map.push(binding));
         EnvironmentMap { map }
     }
 
@@ -139,7 +163,7 @@ impl Default for EnvironmentMap {
 /// LexicalEnvironment represents a series of shallow binding slots for
 /// a procedure of which are populated during the evaluation of the lambda
 /// procedure and during procedure application.
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Debug, PartialEq, Eq, Clone)]
 pub struct LexicalEnvironment {
     slots: RefCell<Vec<VCell>>,
 }
@@ -416,6 +440,46 @@ fn find_free_symbols_in_proc<'a>(
     Ok(())
 }
 
+/// Interally defined symbols
+///
+/// Given the body of a lambda, return the list of internally defined
+/// symbols as a result of calls to define at the beginning of the block.
+///
+/// Internal define expressions are only allowed at the beginning of the
+/// block. Any additional defines will be skipped in anticipation of the
+/// compiler compiling this lambda expression forming an error.
+///
+/// This function also acts as a syntax check to ensure define is only
+/// present at the beginning of a body.
+///
+/// # Arguments
+/// `body` - The body of the lambda
+pub fn internally_defined_symbols(body: &Cell) -> Result<HashSet<&Cell>, Error> {
+    let mut symbols = HashSet::new();
+    let mut beginning_of_body = true;
+
+    for expr in body {
+        if expr.is_pair() && expr.car().unwrap().is_define() {
+            if !beginning_of_body {
+                return Err(InvalidDefineSyntax(format!("out of context: {}", expr)));
+            }
+            let expr = expr.cdr().unwrap();
+            if expr.is_pair() {
+                let expr = expr.car().unwrap();
+                if expr.is_symbol() {
+                    symbols.insert(expr);
+                } else if expr.is_pair() && expr.car().unwrap().is_symbol() {
+                    symbols.insert(expr.car().unwrap());
+                }
+            }
+            continue;
+        } else {
+            beginning_of_body = false;
+        }
+    }
+    Ok(symbols)
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -445,7 +509,7 @@ mod tests {
 
     #[test]
     fn primitive_symbol_check() {
-        assert!(cell!["+"].is_primitive_symbol());
+        assert!(cell!["lambda"].is_primitive_symbol());
         assert!(!cell!["foo"].is_primitive_symbol());
         assert!(!cell![100].is_primitive_symbol());
     }
@@ -472,6 +536,8 @@ mod tests {
         assert_eq!(
             free_symbols(&parse!["(+ (* a b) (* c d) e)"]),
             Ok(HashSet::from([
+                &cell!["+"],
+                &cell!["*"],
                 &cell!["a"],
                 &cell!["b"],
                 &cell!["c"],
@@ -495,12 +561,12 @@ mod tests {
         // lambdas
         assert_eq!(
             free_symbols(&parse!["(lambda (x) (+ x y) z)"]),
-            Ok(HashSet::from([&cell!["y"], &cell!["z"]]))
+            Ok(HashSet::from([&cell!["y"], &cell!["z"], &cell!["+"]]))
         );
 
         assert_eq!(
             free_symbols(&parse!("(lambda (n) (+ ((adder num) n)))")),
-            Ok(HashSet::from([&cell!["adder"], &cell!["num"]]))
+            Ok(HashSet::from([&cell!["adder"], &cell!["num"], &cell!["+"]]))
         );
 
         assert!(free_symbols(&parse!["(lambda)"]).is_err());
@@ -512,4 +578,16 @@ mod tests {
     fn free_syms_returns_errors() {
         assert!(free_symbols(&parse!["(define)"]).is_err());
     }
+
+    #[test]
+    fn internally_defined_symbols_returns_vec() {
+        assert_eq!(
+            internally_defined_symbols(&parse!["((define foo 10)(define (bar baz) 10))"]),
+            Ok(HashSet::from([&cell!["foo"], &cell!["bar"]]))
+        );
+        assert!(internally_defined_symbols(&parse![
+            "((define foo 10)(set! foo 5)(define (bar baz) 10))"
+        ])
+        .is_err());
+    }
 }
