@@ -1,5 +1,6 @@
 use crate::cell;
 use crate::cell::Cell;
+use crate::vm::continuation::Continuation;
 use crate::vm::gc;
 use crate::vm::gc::State;
 use crate::vm::lambda::Lambda;
@@ -13,7 +14,7 @@ pub struct Heap {
     free_list: Vec<usize>,
     heap: Vec<VCell>,
     heap_map: gc::Map,
-    str_map: HashMap<String, usize>,
+    symbol_table: HashMap<String, usize>,
 }
 
 impl Heap {
@@ -26,7 +27,7 @@ impl Heap {
             heap: vec![VCell::undefined(); chunk_size],
             free_list: (0..chunk_size).rev().into_iter().collect(),
             heap_map: gc::Map::new(chunk_size),
-            str_map: HashMap::new(),
+            symbol_table: HashMap::new(),
         }
     }
 
@@ -46,7 +47,7 @@ impl Heap {
     pub fn free(&mut self, ptr: usize) {
         self.heap_map.set(ptr, State::Free);
         if let Some(VCell::Symbol(sym)) = self.heap.get(ptr) {
-            self.str_map.remove(&**sym);
+            self.symbol_table.remove(&**sym);
         }
         *self.heap.get_mut(ptr).unwrap() = VCell::Undefined;
         self.free_list.push(ptr);
@@ -60,12 +61,12 @@ impl Heap {
         let vcell = vcell.into();
         match &vcell {
             VCell::Ptr(_) => panic!("put() on {} would double box", vcell),
-            VCell::Symbol(sym) => match self.str_map.get(sym.deref()) {
+            VCell::Symbol(sym) => match self.symbol_table.get(sym.deref()) {
                 Some(ptr) => VCell::ptr(*ptr),
                 None => {
                     let ptr = self.alloc();
                     *self.heap.get_mut(ptr).expect("heap index is out of bounds") = vcell.clone();
-                    self.str_map.insert(sym.deref().into(), ptr);
+                    self.symbol_table.insert(sym.deref().into(), ptr);
                     VCell::ptr(ptr)
                 }
             },
@@ -90,18 +91,28 @@ impl Heap {
             cell::Cell::Undefined => self.put(VCell::Undefined),
             cell::Cell::Void => self.put(VCell::Void),
             cell::Cell::Nil => self.put(VCell::Nil),
-            cell::Cell::Number(val) => self.put(VCell::FixedNum(val)),
+            cell::Cell::Number(ref val) => self.put(VCell::Number(val.clone())),
             cell::Cell::Bool(val) => self.put(VCell::Bool(val)),
+            cell::Cell::Char(val) => self.put(VCell::Char(val)),
             cell::Cell::Pair(ref car, ref cdr) => {
                 match (self.put_cell(car.deref()), self.put_cell(cdr.deref())) {
                     (VCell::Ptr(car), VCell::Ptr(cdr)) => self.put(VCell::Pair(car, cdr)),
                     _ => panic!("expected ptr, got {:?}", ast),
                 }
             }
+            cell::Cell::String(ref s) => self.put(VCell::string(s.clone())),
             cell::Cell::Symbol(ref sym) => self.put(VCell::symbol(sym.clone())),
+            cell::Cell::Continuation => panic!("unexpected continuation"),
             cell::Cell::Closure => panic!("unexpected closure"),
             cell::Cell::Macro => panic!("unexpected macro"),
             cell::Cell::Lambda => panic!("unexpected lambda"),
+            cell::Cell::Vector(ref vector) => {
+                let mut outv = Vec::with_capacity(vector.len());
+                for it in vector {
+                    outv.push(self.put_cell(it))
+                }
+                self.put(VCell::vector(outv))
+            }
         }
     }
 
@@ -141,14 +152,14 @@ impl Heap {
 
     /// Get Sym Ref
     ///
-    /// Return a Some(VCell) ptr to the given sym reference if sym is interned, otherwise
+    /// Return a Some(VCell) ptr to the given sym reference if sym is known, otherwise
     /// None.
     ///
     /// # Arguments
     /// `sym` - The symbol to lookup
     pub fn get_sym_ref(&self, sym: &Cell) -> Option<VCell> {
         if let Cell::Symbol(sym) = sym {
-            self.str_map.get(sym).map(|it| VCell::ptr(*it))
+            self.symbol_table.get(sym).map(|it| VCell::ptr(*it))
         } else {
             None
         }
@@ -166,25 +177,37 @@ impl Heap {
     pub fn get_as_cell(&self, vcell: &VCell) -> Cell {
         match vcell {
             VCell::Bool(val) => Cell::Bool(*val),
-            VCell::FixedNum(val) => Cell::Number(*val),
+            VCell::Char(val) => Cell::Char(*val),
+            VCell::Number(val) => Cell::Number(val.clone()),
             VCell::Nil => Cell::Nil,
             VCell::Pair(ref car, cdr) => Cell::new_pair(
                 self.get_as_cell(&VCell::Ptr(*car)),
                 self.get_as_cell(&VCell::Ptr(*cdr)),
             ),
             VCell::Ptr(ptr) => self.get_as_cell(self.get_at_index(*ptr)),
+            VCell::String(s) => Cell::String(s.borrow().deref().into()),
             VCell::Symbol(s) => Cell::Symbol(s.deref().into()),
             VCell::Undefined => Cell::Undefined,
             VCell::Void => Cell::Void,
+            VCell::Continuation(_) => Cell::Continuation,
             VCell::Closure(_, _) => Cell::Closure,
             VCell::Lambda(_) => Cell::Lambda,
             VCell::BuiltInProc(_) => Cell::Lambda,
             VCell::Macro(_) => Cell::Macro,
+            VCell::Vector(vector) => {
+                let mut outv = Vec::with_capacity(vector.len());
+                for idx in 0..vector.len() {
+                    outv.push(self.get_as_cell(&vector.get(idx).unwrap()));
+                }
+                Cell::Vector(outv)
+            }
             // Any internal values used by bytecode aren't convertible to Cells and
             // result in a panic.
             VCell::Acc
+            | VCell::ArgumentCount(_)
             | VCell::BasePointer(_)
             | VCell::BasePointerOffset(_)
+            | VCell::EnvironmentPointer(_)
             | VCell::GlobalEnvSlot(_)
             | VCell::LexicalEnv(_)
             | VCell::LexicalEnvSlot(_)
@@ -220,7 +243,7 @@ impl Heap {
                 self.heap_map.mark(ptr);
             }
 
-            trace!("mark {} => {}", ptr, vcell);
+            //trace!("mark {} => {}", ptr, vcell);
             match vcell {
                 VCell::Pair(car, cdr) => {
                     self.mark(car);
@@ -229,6 +252,9 @@ impl Heap {
                 VCell::Ptr(cdr) => {
                     ptr = cdr;
                 }
+                VCell::Continuation(cont) => {
+                    self.mark_continuation(&*cont);
+                }
                 VCell::Lambda(ptr) => {
                     self.mark_lambda(&*ptr);
                 }
@@ -242,18 +268,28 @@ impl Heap {
                         self.mark_vcell(&env.get(it));
                     }
                 }
+                VCell::Vector(vector) => {
+                    for idx in 0..vector.len() {
+                        let vcell = vector.get(idx).unwrap();
+                        self.mark_vcell(&vcell);
+                    }
+                }
+                VCell::EnvironmentPointer(ptr) => self.mark(ptr),
                 VCell::Acc
+                | VCell::ArgumentCount(_)
                 | VCell::BasePointer(_)
                 | VCell::BasePointerOffset(_)
                 | VCell::Bool(_)
+                | VCell::Char(_)
                 | VCell::BuiltInProc(_)
                 | VCell::GlobalEnvSlot(_)
                 | VCell::LexicalEnvSlot(_)
                 | VCell::LexicalEnvPtr(_, _)
-                | VCell::FixedNum(_)
                 | VCell::InstructionPointer(_, _)
                 | VCell::Nil
+                | VCell::Number(_)
                 | VCell::OpCode(_)
+                | VCell::String(_)
                 | VCell::Symbol(_)
                 | VCell::Macro(_)
                 | VCell::Undefined
@@ -267,6 +303,9 @@ impl Heap {
             VCell::InstructionPointer(lambda, _) => {
                 self.mark(*lambda);
             }
+            VCell::Continuation(cont) => {
+                self.mark_continuation(&*cont);
+            }
             VCell::Lambda(lambda) => self.mark_lambda(lambda.as_ref()),
             VCell::Closure(lambda, env) => {
                 self.mark(*lambda);
@@ -282,16 +321,26 @@ impl Heap {
             VCell::LexicalEnvPtr(ptr, _) => {
                 self.mark(*ptr);
             }
+            VCell::Vector(vector) => {
+                for idx in 0..vector.len() {
+                    let vcell = vector.get(idx).unwrap();
+                    self.mark_vcell(&vcell);
+                }
+            }
+            VCell::EnvironmentPointer(ep) => self.mark(*ep),
             VCell::Acc
+            | VCell::ArgumentCount(_)
             | VCell::BasePointer(_)
             | VCell::BasePointerOffset(_)
             | VCell::Bool(_)
+            | VCell::Char(_)
             | VCell::GlobalEnvSlot(_)
-            | VCell::FixedNum(_)
             | VCell::LexicalEnv(_)
             | VCell::LexicalEnvSlot(_)
             | VCell::Nil
+            | VCell::Number(_)
             | VCell::OpCode(_)
+            | VCell::String(_)
             | VCell::Symbol(_)
             | VCell::BuiltInProc(_)
             | VCell::Macro(_)
@@ -300,6 +349,17 @@ impl Heap {
         }
     }
 
+    /// Mark Continuation
+    ///
+    /// Iterate the saved VM state in the continuation
+    pub fn mark_continuation(&mut self, cont: &Continuation) {
+        for it in cont.stack().iter() {
+            self.mark_vcell(it);
+        }
+        self.mark(cont.ip().0);
+        self.mark(cont.ep());
+    }
+
     /// Mark Lambda
     ///
     /// Iterate the lambda byte code and mark any value that contains a reference type
@@ -343,13 +403,36 @@ impl Heap {
         }
         trace!("freed {} vcell(s)", self.free_list.len() - before);
     }
+
+    /// Size
+    ///
+    /// The total capacity of the heap.
+    pub fn capacity(&self) -> usize {
+        self.heap.len()
+    }
+
+    /// Free Size
+    ///
+    /// The number of nodes in the free list.
+    pub fn free_size(&self) -> usize {
+        self.free_list.len()
+    }
+
+    /// Used Size
+    ///
+    /// The number of nodes in used.
+    pub fn used_size(&self) -> usize {
+        self.capacity() - self.free_size()
+    }
 }
 
 #[cfg(test)]
 mod tests {
     use super::*;
     use crate::cell::Cell;
+    use crate::number::Number;
     use crate::{cell, cons};
+
     const CHUNK_SIZE: usize = 1024;
 
     #[test]
@@ -361,10 +444,10 @@ mod tests {
         assert_eq!(heap.heap_map.get(0), Some(State::Allocated));
         assert_eq!(heap.alloc(), 1);
         assert_eq!(heap.heap_map.get(1), Some(State::Allocated));
-        *heap.get_at_index_mut(0) = VCell::FixedNum(42.into());
-        *heap.get_at_index_mut(1) = VCell::FixedNum(43.into());
-        assert_eq!(heap.get_at_index(0), &VCell::FixedNum(42.into()));
-        assert_eq!(heap.get_at_index(1), &VCell::FixedNum(43.into()));
+        *heap.get_at_index_mut(0) = VCell::Number(Number::from(42));
+        *heap.get_at_index_mut(1) = VCell::Number(Number::from(43));
+        assert_eq!(heap.get_at_index(0), &VCell::Number(Number::from(42)));
+        assert_eq!(heap.get_at_index(1), &VCell::Number(Number::from(43)));
     }
 
     #[test]
