@@ -2,9 +2,10 @@ use crate::cell::Cell;
 use crate::vm::environment::{BindingSource, EnvironmentMap, LexicalEnvironment};
 use crate::vm::lambda::Lambda;
 use crate::vm::opcode::OpCode;
+use crate::vm::vcell::VCell::LexicalEnvPtr;
 use crate::vm::vcell::{BuiltInProc, VCell};
 use crate::vm::Error::{
-    InvalidArgs, InvalidBytecode, InvalidNumArgs, InvalidProcedure, VariableNotBound,
+    InvalidBytecode, InvalidNumArgs, InvalidProcedure, InvalidSyntax, VariableNotBound,
 };
 use crate::vm::{Error, Vm};
 use log::trace;
@@ -16,13 +17,20 @@ impl Vm {
     /// Run the virtual machine until it encounters a HALT instruction,
     /// and return the value contained within the ACC register as a Cell
     pub fn run(&mut self) -> Result<Cell, Error> {
+        self.run_count(usize::MAX).map(|it| it.unwrap())
+    }
+
+    pub fn run_count(&mut self, count: usize) -> Result<Option<Cell>, Error> {
         let mut cycles = 0;
-        self.stack.clear();
         loop {
             cycles += 1;
             if cycles % 256 == 0 {
                 self.run_gc();
             }
+            if cycles == count {
+                self.run_gc();
+                return Ok(None);
+            }
             match self.run_one() {
                 Ok(true) => break,
                 Ok(false) => continue,
@@ -31,8 +39,9 @@ impl Vm {
         }
         trace!("cycles: {}", cycles);
         let cell = self.heap.get_as_cell(&self.acc);
+        self.stack.clear();
         self.run_gc();
-        Ok(cell)
+        Ok(Some(cell))
     }
 
     /// Run One
@@ -94,7 +103,7 @@ impl Vm {
                 let lambda = self.heap.get_at_index(lambda_ptr).as_lambda()?;
 
                 // Build the lexical environment
-                let lexical_env = self.build_lexical_environment(&lambda.envmap)?;
+                let lexical_env = self.build_closure_environment(&lambda.envmap)?;
                 let lexical_env = VCell::LexicalEnv(Rc::new(lexical_env));
                 let lexical_env_ptr = self.heap.put(lexical_env).as_ptr()?;
 
@@ -107,14 +116,27 @@ impl Vm {
                     VCell::Closure(lambda, _) => lambda,
                     VCell::Lambda(_) => self.acc.as_ptr()?,
                     VCell::BuiltInProc(BuiltInProc(func)) => {
-                        func(self)?;
+                        self.acc = match func(self)? {
+                            VCell::Ptr(ptr) => VCell::Ptr(ptr),
+                            vcell => self.heap.put(vcell),
+                        };
+                        return Ok(false);
+                    }
+                    VCell::Continuation(cont) => {
+                        let cont = &*cont;
+                        if self.stack.pop()?.as_argc()? == 0 {
+                            return Err(InvalidSyntax("expected value".into()));
+                        }
+                        let result = self.stack.pop()?.clone();
+                        self.restore_continuation(cont);
+                        self.acc = result;
                         return Ok(false);
                     }
                     other => {
                         return Err(InvalidProcedure(self.heap.get_as_cell(&other).to_string()));
                     }
                 };
-                self.stack.push(VCell::Ptr(self.ep));
+                self.stack.push(VCell::EnvironmentPointer(self.ep));
                 self.stack
                     .push(VCell::InstructionPointer(self.ip.0, self.ip.1));
                 self.ip.0 = lambda;
@@ -125,7 +147,20 @@ impl Vm {
                     VCell::Closure(lambda, _) => lambda,
                     VCell::Lambda(_) => self.acc.as_ptr()?,
                     VCell::BuiltInProc(BuiltInProc(func)) => {
-                        func(self)?;
+                        self.acc = match func(self)? {
+                            VCell::Ptr(ptr) => VCell::Ptr(ptr),
+                            vcell => self.heap.put(vcell),
+                        };
+                        return Ok(false);
+                    }
+                    VCell::Continuation(cont) => {
+                        let cont = &*cont;
+                        if self.stack.pop()?.as_argc()? == 0 {
+                            return Err(InvalidSyntax("expected value".into()));
+                        }
+                        let result = self.stack.pop()?.clone();
+                        self.restore_continuation(cont);
+                        self.acc = result;
                         return Ok(false);
                     }
                     other => {
@@ -133,8 +168,8 @@ impl Vm {
                     }
                 };
 
-                let argc = self.stack.get_offset(0)?.as_fixed_num()? as usize;
-                let frame_argc = self.stack.get(self.bp + 1)?.as_fixed_num()? as usize;
+                let argc = self.stack.get_offset(0)?.as_argc()?;
+                let frame_argc = self.stack.get(self.bp + 1)?.as_argc()?;
 
                 if argc == frame_argc {
                     let saved_bp = self.stack.get(self.bp + 4)?.clone();
@@ -158,7 +193,7 @@ impl Vm {
                         self.stack.push(val);
                     }
 
-                    self.stack.push(VCell::FixedNum(argc as i64));
+                    self.stack.push(VCell::ArgumentCount(argc));
                     self.stack.push(saved_ep);
                     self.stack.push(saved_ip);
                     self.bp = saved_bp.as_bp()?;
@@ -167,28 +202,39 @@ impl Vm {
                 }
             }
             OpCode::Enter => {
-                let (lambda, lexical_env) = match self.heap.get(&self.acc) {
-                    VCell::Closure(lambda, lexical_env) => (lambda, lexical_env),
-                    VCell::Lambda(_) => (self.acc.as_ptr()?, self.ep),
+                let (lambda, closure_env) = match self.heap.get(&self.acc) {
+                    VCell::Closure(lambda, lexical_env) => (lambda, Some(lexical_env)),
+                    VCell::Lambda(_) => (self.acc.as_ptr()?, None),
                     _ => {
                         return Err(InvalidBytecode);
                     }
                 };
+
                 let lambda = self.heap.get_at_index(lambda);
                 let lambda = lambda.as_lambda()?;
-                if self.stack.get_offset(-2)?.as_fixed_num()? as usize != lambda.args.len() {
+                if self.stack.get_offset(-2)?.as_argc()? != lambda.args.len() {
                     return Err(InvalidNumArgs("procedure".into()));
                 }
 
                 self.stack.push(VCell::BasePointer(self.bp));
                 self.bp = self.stack.get_sp() - 4;
-                self.ep = lexical_env;
+
+                if let Some(closure_env_ptr) = closure_env {
+                    let closure_env = self.heap.get_at_index(closure_env_ptr).as_lexical_env()?;
+                    let lexical_env =
+                        self.build_lexical_environment(lambda, closure_env_ptr, closure_env)?;
+                    let lexical_env_ptr = self
+                        .heap
+                        .put(VCell::LexicalEnv(Rc::new(lexical_env)))
+                        .as_ptr()?;
+                    self.ep = lexical_env_ptr;
+                }
             }
             OpCode::Ret => {
-                let n = self.stack.get(self.bp + 1)?.as_fixed_num()? as usize;
+                let n = self.stack.get(self.bp + 1)?.as_argc()?;
 
                 *self.stack.get_sp_mut() = self.bp - n;
-                self.ep = self.stack.get(self.bp + 2)?.as_ptr()?;
+                self.ep = self.stack.get(self.bp + 2)?.as_ep()?;
                 self.ip = self.stack.get(self.bp + 3)?.as_ip()?;
                 self.bp = self.stack.get(self.bp + 4)?.as_bp()?;
             }
@@ -197,7 +243,7 @@ impl Vm {
                 // This requires popping off every optional argument, forming a list on the heap,
                 // and placing the list at the top of the argument stack.
                 let req_argc = self.lambda().args.len() - 1;
-                let argc = self.stack.get_offset(-2)?.as_fixed_num()? as usize;
+                let argc = self.stack.get_offset(-2)?.as_argc()?;
                 if argc < req_argc {
                     return Err(InvalidNumArgs("procedure".into()));
                 }
@@ -226,47 +272,11 @@ impl Vm {
                     // Push the list on the stack, a new argc, and restore the caller's
                     // frame
                     self.stack.push(VCell::ptr(varargs));
-                    self.stack.push(VCell::FixedNum((req_argc + 1) as i64));
+                    self.stack.push(VCell::ArgumentCount(req_argc + 1));
                     self.stack.push(saved_ip);
                     self.stack.push(saved_ep);
                 }
             }
-            // Numbers
-            //
-            // The opcodes in this section provide primitives for monipulating numbers,
-            // including primitives for addition, substraction, etc.
-            //
-            OpCode::Add | OpCode::Sub | OpCode::Mul => {
-                let proc_name = match op_code {
-                    OpCode::Add => "+",
-                    OpCode::Sub => "-",
-                    OpCode::Mul => "*",
-                    _ => return Err(InvalidBytecode),
-                };
-                let x = self.heap.get(&self.acc);
-                let y = self.pop_stack()?;
-                let y = self.heap.get(&y);
-                let x = x.as_fixed_num().map_err(|_| {
-                    InvalidArgs(
-                        proc_name.to_string(),
-                        "number".to_string(),
-                        self.heap.get_as_cell(&x).to_string(),
-                    )
-                })?;
-                let y = y.as_fixed_num().map_err(|_| {
-                    InvalidArgs(
-                        proc_name.to_string(),
-                        "number".to_string(),
-                        self.heap.get_as_cell(&y).to_string(),
-                    )
-                })?;
-                match op_code {
-                    OpCode::Add => self.acc = self.heap.put(VCell::fixed_num(x + y)),
-                    OpCode::Sub => self.acc = self.heap.put(VCell::fixed_num(y - x)),
-                    OpCode::Mul => self.acc = self.heap.put(VCell::fixed_num(x * y)),
-                    _ => return Err(InvalidBytecode),
-                };
-            }
         }
 
         Ok(false)
@@ -304,13 +314,6 @@ impl Vm {
         }
     }
 
-    /// Pop Stack
-    ///
-    /// Pop a value off the stack. Error if the stack is empty.
-    fn pop_stack(&mut self) -> Result<VCell, Error> {
-        self.stack.pop().map(|it| it.clone())
-    }
-
     /// Lambda
     ///
     /// Dereference the currently executing lambda
@@ -383,20 +386,20 @@ impl Vm {
             VCell::GlobalEnvSlot(slot) => {
                 self.globenv.put_slot(slot, vcell);
             }
-            VCell::LexicalEnvSlot(slot) => match self.heap.get_at_index(self.ep) {
-                VCell::LexicalEnv(env) => {
-                    env.put(slot, vcell);
+            VCell::LexicalEnvSlot(slot) => {
+                let lexical_env = self.heap.get_at_index(self.ep).as_lexical_env()?;
+                match lexical_env.get(slot) {
+                    VCell::LexicalEnvPtr(env, slot) => {
+                        self.heap
+                            .get_at_index(env)
+                            .as_lexical_env()?
+                            .put(slot, vcell);
+                    }
+                    _ => {
+                        lexical_env.put(slot, vcell);
+                    }
                 }
-                VCell::LexicalEnvPtr(env, slot) => {
-                    self.heap
-                        .get_at_index(*env)
-                        .as_lexical_env()?
-                        .put(*slot, vcell);
-                }
-                _ => {
-                    return Err(InvalidBytecode);
-                }
-            },
+            }
             _ => return Err(InvalidBytecode),
         }
         Ok(())
@@ -423,7 +426,7 @@ impl Vm {
     /// Load an argument from the current stack frame given the argument
     /// index.
     fn load_arg(&self, index: usize) -> Result<&VCell, Error> {
-        let arg_count = self.stack.get(self.bp + 1)?.as_fixed_num()? as usize;
+        let arg_count = self.stack.get(self.bp + 1)?.as_argc()?;
         let offset = self.bp;
         let offset = (offset - arg_count) + index + 1;
         self.stack.get(offset)
@@ -433,12 +436,18 @@ impl Vm {
     ///
     /// Run GC performs two steps in order:
     ///
-    /// 1. It performs a mark on all roots:
+    /// 1. Check if heap utilization is > 75%, aborting gc is not.
+    ///
+    /// 2. It performs a mark on all roots:
     ///    * The global environment
     ///    * Any data referecned by the running program & stack
     ///
-    /// 2. A sweep, freeing any vcells not marked as used in step #1.
+    /// 3. A sweep, freeing any vcells not marked as used in step #1.
     pub fn run_gc(&mut self) {
+        if (self.heap.used_size() as f64 / self.heap.capacity() as f64) < 0.75_f64 {
+            return;
+        }
+
         self.globenv
             .iter_bindings()
             .for_each(|it| self.heap.mark(*it));
@@ -448,20 +457,24 @@ impl Vm {
             .filter_map(|it| it.as_ptr().ok())
             .for_each(|it| self.heap.mark(it));
 
-        self.stack.iter().for_each(|it| self.heap.mark_vcell(it));
+        self.stack
+            .iter_to_sp()
+            .for_each(|it| self.heap.mark_vcell(it));
         self.heap.mark_vcell(&self.acc);
         self.heap.mark(self.ip.0);
+        self.heap.mark(self.ep);
         self.heap.sweep();
     }
 
-    /// Build Lexical Environment
+    /// Build Closure Environment
     ///
     /// Build a lexical environment with the given environment map, assuming
-    /// that the current stack is that of the IOF.
+    /// that the current stack is that of the IOF. This function is called by
+    /// the CLOSURE instruction as part of creation of a lambda.
     ///
     /// # Arguments
     /// `envmap` - The EnvironmentMap used to build the lexical environment
-    fn build_lexical_environment(
+    fn build_closure_environment(
         &self,
         envmap: &EnvironmentMap,
     ) -> Result<LexicalEnvironment, Error> {
@@ -472,11 +485,52 @@ impl Vm {
                     lexical_env.put(slot, self.load_arg(arg)?.clone());
                 }
                 BindingSource::IofEnvironment(iof_slot) => {
-                    lexical_env.put(slot, VCell::LexicalEnvPtr(self.ep, iof_slot));
+                    let iof_env = self.heap.get_at_index(self.ep).as_lexical_env()?;
+                    match iof_env.get(iof_slot) {
+                        VCell::LexicalEnvPtr(_, _) => lexical_env.put(slot, iof_env.get(iof_slot)),
+                        _ => lexical_env.put(slot, VCell::LexicalEnvPtr(self.ep, iof_slot)),
+                    };
                 }
-                // Own arguments can't be captured in the environment until the procedure
-                // is applied.
-                BindingSource::Global | BindingSource::Argument(_) => {}
+                // Argument bindings aren't available until CALL/TCALL, and internal definitions are not
+                // defined until the lambda's body executes.
+                BindingSource::Global
+                | BindingSource::Argument(_)
+                | BindingSource::InternalDefinition => {}
+            }
+        }
+        Ok(lexical_env)
+    }
+
+    /// Chain Lexical Environment
+    ///
+    /// Given a closure's environment, produce a cloned environment that chains
+    /// to the original environment. This function is called by the ENTER instruction
+    /// when applying a function.
+    ///
+    /// # Arguments
+    /// `lambda` - The lambda being applied by the ENTER instruction.
+    /// `env_ptr` - A pointer to the base environment of lambda.
+    /// `env` - The base environment.
+    fn build_lexical_environment(
+        &self,
+        lambda: &Lambda,
+        closure_env_ptr: usize,
+        closure_env: &LexicalEnvironment,
+    ) -> Result<LexicalEnvironment, Error> {
+        let lexical_env = closure_env.clone();
+        for (slot, it) in lambda.envmap.get_map().iter().enumerate() {
+            match it.1 {
+                BindingSource::Argument(arg) => {
+                    let arg_offset = self.bp - (lambda.argc() - arg) + 1;
+                    lexical_env.put(slot, self.stack.get(arg_offset)?.clone());
+                }
+                BindingSource::IofArgument(_) | BindingSource::IofEnvironment(_) => {
+                    match closure_env.get(slot) {
+                        VCell::LexicalEnvPtr(_, _) => {}
+                        _ => lexical_env.put(slot, LexicalEnvPtr(closure_env_ptr, slot)),
+                    }
+                }
+                BindingSource::Global | BindingSource::InternalDefinition => {}
             }
         }
         Ok(lexical_env)
@@ -484,16 +538,20 @@ impl Vm {
 
     #[allow(dead_code)]
     fn trace_environment_map(&self, envmap: &EnvironmentMap) {
+        trace!("--- ENVIRONMENT MAP ---");
         envmap.get_map().iter().for_each(|it| {
             trace!("{} => {:?}", self.get_str_bound_to(it.0.clone()), it.1);
-        })
+        });
+        trace!("--- END MAP ---");
     }
 
     #[allow(dead_code)]
     fn trace_lexical_environment(&self, env: &LexicalEnvironment) {
+        trace!("--- LEXICAL ENVIRONMENT ---");
         for it in 0..env.slot_len() {
-            trace!("{} => {}", it, env.get(it));
+            trace!("{} => {:?}", it, env.get(it));
         }
+        trace!("--- END LEXICAL ENVIRONMENT ---");
     }
 
     fn trace_instruction(&self) {
@@ -514,7 +572,7 @@ impl Vm {
                 VCell::InstructionPointer(self.ip.0, self.ip.1),
                 VCell::Ptr(self.stack.get_sp()),
                 VCell::BasePointer(self.bp),
-                VCell::Ptr(self.ep)
+                VCell::EnvironmentPointer(self.ep)
             )
         );
     }
