@@ -49,6 +49,13 @@ impl Stack {
         self.stack.iter()
     }
 
+    /// Iter
+    ///
+    /// Return an iterator to the stack vector up to the stack pointer
+    pub fn iter_to_sp(&self) -> impl Iterator<Item = &VCell> {
+        self.stack[0..self.sp + 1].iter()
+    }
+
     /// Get
     ///
     /// Get a VC at the given stack location.
@@ -161,6 +168,21 @@ impl Stack {
             );
         }
     }
+
+    pub fn to_continuation(&self) -> Stack {
+        Stack {
+            stack: self.stack[0..self.sp + 1].to_vec(),
+            sp: self.sp,
+        }
+    }
+
+    pub fn restore_continuation(&mut self, cont: &Stack) {
+        self.stack
+            .split_at_mut(cont.stack.len())
+            .0
+            .clone_from_slice(&cont.stack);
+        self.sp = cont.sp;
+    }
 }
 
 impl Default for Stack {
@@ -172,13 +194,14 @@ impl Default for Stack {
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::number::Number;
 
     #[test]
     fn stack_grows_on_push() {
         let mut stack = Stack::new();
         assert_eq!(stack.len(), 256);
         for i in 0..1024 {
-            stack.push(VCell::FixedNum(i));
+            stack.push(VCell::Number(Number::from(i)));
         }
         assert_eq!(stack.len(), 2048)
     }
@@ -186,24 +209,47 @@ mod tests {
     #[test]
     fn relative_access() {
         let mut stack = Stack::new();
-        stack.push(VCell::FixedNum(0));
-        stack.push(VCell::FixedNum(1));
-        stack.push(VCell::FixedNum(2));
+        stack.push(VCell::number(0));
+        stack.push(VCell::number(1));
+        stack.push(VCell::number(2));
         assert_eq!(stack.get_offset(2), Ok(&VCell::Undefined));
         assert_eq!(stack.get_offset(1), Ok(&VCell::Undefined));
-        assert_eq!(stack.get_offset(0), Ok(&VCell::FixedNum(2)));
-        assert_eq!(stack.get_offset(-1), Ok(&VCell::FixedNum(1)));
-        assert_eq!(stack.get_offset(-2), Ok(&VCell::FixedNum(0)));
+        assert_eq!(stack.get_offset(0), Ok(&VCell::number(2)));
+        assert_eq!(stack.get_offset(-1), Ok(&VCell::number(1)));
+        assert_eq!(stack.get_offset(-2), Ok(&VCell::number(0)));
         assert_eq!(stack.get_offset(-3), Ok(&VCell::Undefined));
     }
 
     #[test]
     fn push_and_pop() {
         let mut stack = Stack::new();
-        stack.push(VCell::FixedNum(1));
-        stack.push(VCell::FixedNum(2));
-        assert_eq!(stack.pop(), Ok(&VCell::FixedNum(2)));
-        assert_eq!(stack.pop(), Ok(&VCell::FixedNum(1)));
+        stack.push(VCell::number(1));
+        stack.push(VCell::number(2));
+        assert_eq!(stack.pop(), Ok(&VCell::number(2)));
+        assert_eq!(stack.pop(), Ok(&VCell::number(1)));
         assert_eq!(stack.pop(), Err(InvalidStackIndex(0)));
     }
+
+    #[test]
+    fn continuation() {
+        let mut stack = Stack::new();
+        stack.push(VCell::number(1));
+        stack.push(VCell::number(2));
+        stack.push(VCell::number(3));
+        assert_eq!(stack.get_sp(), 3);
+        let cont = stack.to_continuation();
+
+        *stack.get_sp_mut() = 0;
+        stack.push(VCell::number(4));
+        stack.push(VCell::number(5));
+        assert_eq!(stack.get_sp(), 2);
+        assert_eq!(stack.get_offset(0), Ok(&VCell::number(5)));
+        assert_eq!(stack.get_offset(-1), Ok(&VCell::number(4)));
+
+        stack.restore_continuation(&cont);
+        assert_eq!(stack.get_sp(), 3);
+        assert_eq!(stack.get_offset(0), Ok(&VCell::number(3)));
+        assert_eq!(stack.get_offset(-1), Ok(&VCell::number(2)));
+        assert_eq!(stack.get_offset(-2), Ok(&VCell::number(1)));
+    }
 }
