@@ -19,11 +19,150 @@ macro_rules! cdr {
     }};
 }
 
+/// Pattern
+///
+/// Pattern represents a single syntax-rules pattern, where
+/// `pattern` contains the pattern expression and `variables` is
+/// the set of pattern variables encountered in the pattern.
+
+#[derive(Debug, Eq, PartialEq)]
+pub struct Pattern {
+    expr: Cell,
+    variables: Vec<Cell>,
+    expanded_variables: Vec<Cell>,
+    ellipsis: Cell,
+    literals: Vec<Cell>,
+    underscore: Cell,
+}
+
+impl Pattern {
+    pub fn try_new(expr: &Cell, ellipsis: &Cell, literals: &[Cell]) -> Result<Pattern, Error> {
+        if !expr.is_pair() {
+            return Err(InvalidDefineSyntax("pattern must be a ()".into()));
+        }
+
+        let mut pattern = Pattern {
+            expr: expr.clone(),
+            variables: vec![],
+            expanded_variables: vec![],
+            ellipsis: ellipsis.clone(),
+            literals: literals.to_vec(),
+            underscore: cell!["_"],
+        };
+
+        Self::build(cdr!(expr), &mut pattern)?;
+        Ok(pattern)
+    }
+
+    pub fn is_ellipsis(&self, cell: &Cell) -> bool {
+        *cell == self.ellipsis
+    }
+
+    pub fn is_literal(&self, cell: &Cell) -> bool {
+        self.literals.iter().any(|it| it == cell)
+    }
+
+    pub fn is_variable(&self, cell: &Cell) -> bool {
+        self.variables.iter().any(|it| it == cell)
+    }
+
+    pub fn is_expanded_variable(&self, cell: &Cell) -> bool {
+        self.expanded_variables.iter().any(|it| it == cell)
+    }
+
+    fn is_variable_candidate(&self, cell: &Cell) -> bool {
+        cell.is_symbol()
+            && !self.is_literal(cell)
+            && !self.is_ellipsis(cell)
+            && *cell != self.underscore
+    }
+
+    fn build(expr: &Cell, pattern: &mut Pattern) -> Result<(), Error> {
+        let improper = expr.is_improper_list();
+        let len = expr.len();
+        let mut iter = expr.iter().enumerate().peekable();
+        let mut ellipsis_ct = 0;
+        while let Some((idx, it)) = iter.next() {
+            let ellipsis_next = match iter.peek() {
+                Some((_, cell)) => pattern.is_ellipsis(cell),
+                _ => false,
+            };
+            match it {
+                Cell::Symbol(_) => {
+                    if pattern.is_ellipsis(it) {
+                        if idx == 0 || (idx == len - 1 && improper) {
+                            return Err(InvalidDefineSyntax(format!(
+                                "invalid ellipsis placement in {}",
+                                expr
+                            )));
+                        }
+                        ellipsis_ct += 1;
+                        if ellipsis_ct > 1 {
+                            return Err(InvalidDefineSyntax(format!(
+                                "duplicate ellipsis in {}",
+                                expr
+                            )));
+                        }
+                        continue;
+                    }
+
+                    if pattern.is_variable_candidate(it) {
+                        if pattern.is_variable(it) {
+                            return Err(InvalidDefineSyntax(format!(
+                                "duplicate pattern variable {}",
+                                it
+                            )));
+                        }
+                        pattern.variables.push(it.clone())
+                    } else if ellipsis_next {
+                        return Err(InvalidDefineSyntax(
+                            "ellipsis must follow pattern variable".into(),
+                        ));
+                    }
+
+                    if ellipsis_next {
+                        Self::find_expanded_variables(it, pattern);
+                    }
+                }
+                Cell::Pair(_, _) => {
+                    if ellipsis_next {
+                        Self::find_expanded_variables(it, pattern);
+                    }
+                    Self::build(it, pattern)?;
+                }
+                _ => {}
+            }
+        }
+        Ok(())
+    }
+
+    fn find_expanded_variables(expr: &Cell, pattern: &mut Pattern) {
+        match expr {
+            Cell::Symbol(_) => {
+                if pattern.is_variable_candidate(expr)
+                    && !pattern.expanded_variables.iter().any(|it| it == expr)
+                {
+                    pattern.expanded_variables.push(expr.clone());
+                }
+            }
+            Cell::Pair(_, _) => {
+                for it in expr {
+                    Self::find_expanded_variables(it, pattern);
+                }
+            }
+            _ => {}
+        }
+    }
+}
+
+/// Transform
+///
+/// Transform is a runtime representation of a set of syntax-rules.
 #[derive(Debug, Eq, PartialEq)]
 pub struct Transform {
     keyword: Cell,
     ellipsis: Cell,
-    syntax_rules: Vec<(Cell, Cell)>,
+    syntax_rules: Vec<(Pattern, Cell)>,
     literals: Vec<Cell>,
 }
 
@@ -84,8 +223,8 @@ impl Transform {
         for it in syntax_rules {
             let pattern = car!(it).clone();
             let template = car!(cdr!(it)).clone();
-            Self::check_pattern_syntax(&pattern, &ellipsis, &literals)?;
-            Self::check_template_syntax(&pattern, &ellipsis, &literals)?;
+            let pattern = Pattern::try_new(&pattern, &ellipsis, &literals)?;
+            Self::check_template_syntax(&template, &pattern, &ellipsis)?;
             syntax_rules_vec.push((pattern, template));
         }
 
@@ -108,107 +247,46 @@ impl Transform {
         &self.keyword
     }
 
-    /// Check Pattern Syntax
-    ///
-    /// Check numerous rules
-    ///
-    /// * A variable must not appear more than once in the pattern, unless it is
-    ///   a literal or _.
-    /// * An ellipsis may only appear once in a list, and must be preceded
-    ///   by a pattern variable
-    /// * An ellipsis can be the last element in a list, unless it's an improper
-    ///   list.
-    fn check_pattern_syntax(
-        pattern: &Cell,
-        ellipsis: &Cell,
-        literals: &[Cell],
-    ) -> Result<(), Error> {
-        if !pattern.is_pair() {
-            return Err(InvalidDefineSyntax("pattern must be a ()".into()));
-        }
-        let mut variables = vec![];
-        fn check_pattern_syntax<'a, 'b>(
-            pattern: &'a Cell,
-            ellipsis: &'a Cell,
-            literals: &'a [Cell],
-            variables: &'b mut Vec<&'a Cell>,
-        ) -> Result<(), Error> {
-            if pattern.is_pair() && car!(pattern) == ellipsis {
-                return Err(InvalidDefineSyntax("ellipsis out of place".into()));
-            }
-            let improper = pattern.is_improper_list();
-            let mut ellipsis_in_pattern = false;
-            let mut cur = pattern.iter().peekable();
-            while let Some(pattern) = cur.next() {
-                match pattern {
-                    Cell::Pair(_, _) => {
-                        check_pattern_syntax(pattern, ellipsis, literals, variables)?
-                    }
-                    Cell::Symbol(sym) => {
-                        if literals.iter().any(|it| it == pattern) || sym == "_" {
-                            continue;
-                        }
-
-                        if pattern == ellipsis {
-                            if ellipsis_in_pattern || (improper && cur.peek().is_none()) {
-                                return Err(InvalidDefineSyntax("ellipses out of place".into()));
-                            }
-                            ellipsis_in_pattern = true;
-                            continue;
-                        }
-
-                        // All other identifiers must be variables
-                        if variables.iter().any(|it| *it == pattern) {
-                            return Err(InvalidDefineSyntax(format!(
-                                "the pattern variable {} was used more than once",
-                                pattern
-                            )));
-                        } else {
-                            variables.push(pattern);
-                        }
-                    }
-                    _ => {}
-                }
-            }
-
-            Ok(())
-        }
-        check_pattern_syntax(cdr!(pattern), ellipsis, literals, &mut variables)
-    }
-
     /// Check Template Syntax
+    ///
+    /// * Any symbol preceding an ellipsis must be a pattern variable
+    /// * Pattern variables may not be expanded multiple times (TODO)
+    /// * Like patterns, ellipsis must not be in the tail position of an
+    ///   improper list.
     fn check_template_syntax(
-        _pattern: &Cell,
-        _ellipsis: &Cell,
-        _literals: &[Cell],
+        template: &Cell,
+        pattern: &Pattern,
+        ellipsis: &Cell,
     ) -> Result<(), Error> {
-        Ok(())
-    }
+        let improper = template.is_improper_list();
+        let mut ellipsis_in_pattern = false;
+        let mut iter = template.iter().peekable();
 
-    /// Find Pattern Variables
-    ///
-    /// Given the pattern express, build up a set of all pattern
-    /// variables:
-    ///
-    /// * Variables which are not literals
-    /// * Variables which are not _
-    /// * Variables which are not the ellipsis
-    pub fn find_pattern_variables<'a>(&self, pattern: &'a Cell, variables: &mut Vec<&'a Cell>) {
-        match pattern {
-            Cell::Symbol(_) => {
-                if !self.is_literal(pattern) & !(pattern == &cell!["_"])
-                    && !(pattern == &self.ellipsis)
-                {
-                    variables.push(pattern);
-                }
-            }
-            Cell::Pair(_, _) => {
-                for it in pattern {
-                    self.find_pattern_variables(it, variables);
-                }
-            }
-            _ => {}
+        if template.is_pair() && car!(template) == ellipsis {
+            return Err(InvalidDefineSyntax("ellipsis out of place".into()));
         }
+
+        while let Some(template) = iter.next() {
+            match template {
+                Cell::Pair(_, _) => Self::check_template_syntax(template, pattern, ellipsis)?,
+                Cell::Symbol(_) => {
+                    if !pattern.is_variable(template) && iter.peek() == Some(&ellipsis) {
+                        return Err(InvalidDefineSyntax(
+                            "ellipses must follow a pattern variable".into(),
+                        ));
+                    }
+                    if template == ellipsis {
+                        if ellipsis_in_pattern || (improper && iter.peek().is_none()) {
+                            return Err(InvalidDefineSyntax("ellipses out of place".into()));
+                        }
+                        ellipsis_in_pattern = true;
+                        continue;
+                    }
+                }
+                _ => {}
+            }
+        }
+        Ok(())
     }
 
     /// Transform
@@ -226,18 +304,18 @@ impl Transform {
         }
 
         for rule in &self.syntax_rules {
-            let mut pattern_variables = vec![];
-            self.find_pattern_variables(cdr!(&rule.0), &mut pattern_variables);
-            let mut env = PatternEnvironment::new();
-            env.variables = pattern_variables;
-            if self.pattern_match(cdr!(&rule.0), cdr!(expr), &mut env) {
+            let mut env = PatternEnvironment::new(&rule.0);
+            if self.pattern_match(cdr!(&rule.0.expr), cdr!(expr), &mut env) {
                 return self
-                    .expand(&rule.1, &mut env)
+                    .expand(&rule.1, &rule.0, &mut env)
                     .ok_or_else(|| InvalidSyntax(self.keyword.to_string()));
             }
         }
 
-        return Err(InvalidSyntax(format!("bad use of {}", self.keyword)));
+        return Err(InvalidSyntax(format!(
+            "no matching syntax for {}",
+            self.keyword
+        )));
     }
 
     /// Pattern Match
@@ -261,24 +339,73 @@ impl Transform {
         }
         let mut expr_iter = expr.iter().peekable();
         let mut pattern_iter = pattern.iter().peekable();
-        if pattern_iter.peek().is_none() {
-            return expr_iter.peek().is_none();
-        }
 
-        let mut pattern = pattern_iter.next().unwrap();
+        let mut expr;
+        let mut pattern = &Cell::Nil;
+
+        let mut in_ellipsis = false;
+
         loop {
-            let in_ellipsis = pattern_iter.peek() == Some(&&self.ellipsis);
-
-            let expr = match expr_iter.next() {
+            // Get the next expression
+            // If there is no next expression, then:
+            //   * If we are in an ellipsis expansion, move the pattern iterator
+            //     past it so we can see if more pattern remains.
+            //   * If any pattern remains, there is no match.
+            //   * If pattern was exhausted, then the match is complete.
+            expr = match expr_iter.next() {
                 Some(expr) => expr,
                 None => {
-                    return in_ellipsis;
+                    if in_ellipsis {
+                        pattern_iter.next();
+                    }
+                    return match pattern_iter.next() {
+                        Some(_) => {
+                            if pattern_iter.peek() == Some(&&self.ellipsis) {
+                                pattern_iter.next();
+                                pattern_iter.peek().is_none()
+                            } else {
+                                false
+                            }
+                        }
+                        None => true,
+                    };
                 }
             };
 
+            // Get the next pattern.
+            // * Reuse the same pattern if we're in an ellipsis expansion
+            //   and based on the expression length there's more to capture.
+            pattern = match in_ellipsis {
+                true => {
+                    if pattern_iter.len() == expr_iter.len() + 2 {
+                        pattern_iter.next();
+                        match pattern_iter.next() {
+                            Some(pattern) => pattern,
+                            None => return expr_iter.peek().is_none(),
+                        }
+                    } else {
+                        pattern
+                    }
+                }
+                false => match pattern_iter.next() {
+                    Some(pattern) => pattern,
+                    None => {
+                        return false;
+                    }
+                },
+            };
+
+            in_ellipsis = pattern_iter.peek() == Some(&&self.ellipsis);
+
             match pattern {
                 Cell::Symbol(_) => {
-                    env.add_binding(pattern, expr);
+                    if self.is_literal(pattern) {
+                        if pattern != expr {
+                            return false;
+                        }
+                    } else if pattern != &cell!["_"] {
+                        env.add_binding(pattern, expr);
+                    }
                 }
                 Cell::Pair(_, _) => {
                     if !self.pattern_match(pattern, expr, env) {
@@ -291,27 +418,6 @@ impl Transform {
                     }
                 }
             }
-
-            if in_ellipsis {
-                let pattern_len = pattern_iter.peek().map_or(0, |it| it.len()) - 1;
-                let expr_len = expr_iter.peek().map_or(0, |it| it.len());
-                if expr_len == pattern_len {
-                    pattern_iter.next();
-                    pattern = match pattern_iter.next() {
-                        Some(pattern) => pattern,
-                        None => {
-                            return expr_iter.peek().is_none();
-                        }
-                    }
-                }
-            } else {
-                pattern = match pattern_iter.next() {
-                    Some(pattern) => pattern,
-                    None => {
-                        return expr_iter.peek().is_none();
-                    }
-                }
-            }
         }
     }
 
@@ -322,11 +428,17 @@ impl Transform {
     ///
     /// # Arguments
     /// `template` - The template to use for expansion
+    /// `pattern` - The pattern associated with the template being expanded.
     /// `bindings` The matched bindings from the pattern
-    fn expand(&self, template: &Cell, env: &mut PatternEnvironment) -> Option<Cell> {
+    fn expand(
+        &self,
+        template: &Cell,
+        pattern: &Pattern,
+        env: &mut PatternEnvironment,
+    ) -> Option<Cell> {
         match template {
             Cell::Symbol(_) => {
-                return if env.is_variable(template) {
+                return if pattern.is_variable(template) {
                     env.get_binding(template).cloned()
                 } else {
                     Some(template.clone())
@@ -336,9 +448,10 @@ impl Transform {
                 let mut v = vec![];
                 let mut template_iter = template.iter().peekable();
                 let mut template = template_iter.next().unwrap();
+
                 loop {
                     let in_ellipsis = template_iter.peek() == Some(&&self.ellipsis);
-                    match self.expand(template, env) {
+                    match self.expand(template, pattern, env) {
                         Some(cell) => {
                             v.push(cell);
                             if in_ellipsis {
@@ -372,23 +485,30 @@ impl Transform {
 /// Pattern environment is the result of a successful pattern,
 /// containing all of the information needed to apply the template
 /// portion of the pattern rule.
-#[derive(Debug)]
+#[derive(Debug, Clone)]
 struct PatternEnvironment<'a> {
     /// Bindings are pairs of matched (pattern expr)
     bindings: Vec<(&'a Cell, &'a Cell)>,
 
-    /// Variables are a list of all matching pattern
-    /// variable identifiers and can be used to test if
-    /// an identifier in the template is a pattern variable
-    /// or not.
-    variables: Vec<&'a Cell>,
+    /// A copy of the pattern, which is used to determine what
+    /// type of bindings a variable is
+    pattern: &'a Pattern,
+
+    /// Position of expanded bindings (bindings that were captured
+    /// and expanded with the ellipsis)
+    iters: Vec<(&'a Cell, Option<usize>)>,
 }
 
 impl<'a> PatternEnvironment<'a> {
-    fn new() -> PatternEnvironment<'a> {
+    fn new(pattern: &'a Pattern) -> PatternEnvironment<'a> {
         PatternEnvironment {
             bindings: vec![],
-            variables: vec![],
+            pattern,
+            iters: pattern
+                .expanded_variables
+                .iter()
+                .map(|it| (it, None))
+                .collect(),
         }
     }
 
@@ -396,16 +516,46 @@ impl<'a> PatternEnvironment<'a> {
         self.bindings.push((pattern, expr));
     }
 
-    fn get_binding(&mut self, pattern: &Cell) -> Option<&'a Cell> {
-        if let Some(idx) = self.bindings.iter().position(|it| it.0 == pattern) {
-            Some(self.bindings.remove(idx).1)
-        } else {
+    fn get_binding(&mut self, symbol: &Cell) -> Option<&'a Cell> {
+        if !self.pattern.is_variable(symbol) {
             None
+        } else if self.pattern.is_expanded_variable(symbol) {
+            self.get_expanded_binding(symbol)
+        } else {
+            self.bindings
+                .iter()
+                .find(|it| it.0 == symbol)
+                .map(|it| it.1)
         }
     }
 
-    fn is_variable(&self, expr: &Cell) -> bool {
-        self.variables.iter().any(|it| *it == expr)
+    fn get_expanded_binding(&mut self, symbol: &Cell) -> Option<&'a Cell> {
+        let iter = match self.iters.iter_mut().find(|it| it.0 == symbol) {
+            Some((_, iter)) => iter,
+            None => {
+                return None;
+            }
+        };
+
+        let start = match iter {
+            Some(position) => *position,
+            None => 0_usize,
+        };
+
+        match self.bindings[start..self.bindings.len()]
+            .iter()
+            .enumerate()
+            .find(|it| it.1 .0 == symbol)
+        {
+            Some(binding) => {
+                *iter = Some(start + binding.0 + 1);
+                Some(binding.1 .1)
+            }
+            None => {
+                *iter = None;
+                None
+            }
+        }
     }
 }
 
@@ -415,6 +565,51 @@ mod tests {
     use crate::parse;
     use crate::{cell, lex};
 
+    #[test]
+    fn bad_patterns() {
+        assert!(Pattern::try_new(&parse!("#t"), &cell!["..."], &vec![]).is_err());
+        assert!(Pattern::try_new(&parse!("(_ ... a)"), &cell!["..."], &vec![]).is_err());
+        assert!(Pattern::try_new(&parse!("(_ a . ...)"), &cell!["..."], &vec![]).is_err());
+        assert!(Pattern::try_new(&parse!("(_ a a)"), &cell!["..."], &vec![]).is_err());
+    }
+
+    #[test]
+    fn pattern_variable() {
+        assert_eq!(
+            Pattern::try_new(&parse!("(_ a b c)"), &cell!["..."], &vec![])
+                .unwrap()
+                .variables,
+            vec![cell!["a"], cell!["b"], cell!["c"]]
+        );
+        assert_eq!(
+            Pattern::try_new(&parse!("(_ a b . c)"), &cell!["..."], &vec![])
+                .unwrap()
+                .variables,
+            vec![cell!["a"], cell!["b"], cell!["c"]]
+        );
+        assert_eq!(
+            Pattern::try_new(&parse!("(_ a* ...)"), &cell!["..."], &vec![])
+                .unwrap()
+                .variables,
+            vec![cell!["a*"]]
+        );
+        assert_eq!(
+            Pattern::try_new(&parse!("(_ (a* b*) ...)"), &cell!["..."], &vec![])
+                .unwrap()
+                .variables,
+            vec![cell!["a*"], cell!["b*"]]
+        );
+    }
+
+    #[test]
+    fn expanded_pattern_variables() {
+        let pattern =
+            Pattern::try_new(&parse!("(_ a (b (c ...)) ...)"), &cell!["..."], &vec![]).unwrap();
+        assert_eq!(pattern.variables, vec![cell!["a"], cell!["b"], cell!["c"]]);
+        assert_eq!(pattern.expanded_variables, vec![cell!["b"], cell!["c"]]);
+    }
+
+    #[test]
     #[test]
     fn error_on_bad_form() {
         assert!(Transform::try_new(&parse!("(define-syntax)")).is_err());
@@ -484,6 +679,18 @@ mod tests {
         "#
         ))
         .is_err());
+
+        // ellipsis in head position
+        assert!(Transform::try_new(&parse!(
+            r#"
+        (define-syntax bad
+              (syntax-rules ()
+                [(_ (...)) ()]))
+        "#
+        ))
+        .is_err());
+
+        // Ellipsis in improper list tail
         assert!(Transform::try_new(&parse!(
             r#"
         (define-syntax bad
@@ -492,6 +699,74 @@ mod tests {
         "#
         ))
         .is_err());
+
+        // Ellipsis matching the keyword
+        assert!(Transform::try_new(&parse!(
+            r#"
+        (define-syntax bad
+              (syntax-rules ()
+                [(_ ...) ()]))
+        "#
+        ))
+        .is_err());
+
+        // Ellipsis matching _
+        assert!(Transform::try_new(&parse!(
+            r#"
+        (define-syntax bad
+              (syntax-rules ()
+                [(_ _ ...) ()]))
+        "#
+        ))
+        .is_err());
+
+        // Ellipsis matching a literal
+        assert!(Transform::try_new(&parse!(
+            r#"
+        (define-syntax bad
+              (syntax-rules (literal)
+                [(_ literal ...) ()]))
+        "#
+        ))
+        .is_err());
+    }
+
+    #[test]
+    fn bad_template_syntax() {
+        // Expansion of a non-pattern variavle
+        assert!(Transform::try_new(&parse!(
+            r#"
+        (define-syntax bad
+              (syntax-rules ()
+                [(_ a ...) (b ...)]))
+        "#
+        ))
+        .is_err());
+        // Invalid ellipsis position
+        assert!(Transform::try_new(&parse!(
+            r#"
+        (define-syntax bad
+              (syntax-rules ()
+                [(_ a ...) (...)]))
+        "#
+        ))
+        .is_err());
+        assert!(Transform::try_new(&parse!(
+            r#"
+        (define-syntax bad
+              (syntax-rules ()
+                [(_ a ...) (a ... ...)]))
+        "#
+        ))
+        .is_err());
+        assert!(Transform::try_new(&parse!(
+            r#"
+        (define-syntax bad
+              (syntax-rules ()
+                [(_ a ...) (a . ...)]))
+        "#
+        ))
+        .is_err());
     }
 
     #[test]
@@ -569,7 +844,7 @@ mod tests {
             r#"
         (define-syntax sum
               (syntax-rules ()
-                [(sum *a ...) (+ *a ...)]
+                [(sum a* ...) (+ a* ...)]
         ))
         "#
         ))
@@ -585,6 +860,173 @@ mod tests {
         );
     }
 
+    #[test]
+    fn expansion_edge_cases() {
+        {
+            let transform = Transform::try_new(&parse!(
+                r#"
+            (define-syntax sum
+                  (syntax-rules ()
+                    [(sum a1 a* ... a2) (+ a1 a* ... a2)]
+            ))
+            "#
+            ))
+            .unwrap();
+            assert!(transform.transform(&parse!("(sum 10 20)")).is_err());
+            assert_eq!(
+                transform.transform(&parse!("(sum 10 20 30)")),
+                Ok(parse!("(+ 10 20 30)"))
+            );
+        }
+        {
+            let transform = Transform::try_new(&parse!(
+                r#"
+            (define-syntax sum
+                  (syntax-rules ()
+                    [(sum a1 a* ...) (+ a1 a* ...)]
+            ))
+            "#
+            ))
+            .unwrap();
+            assert_eq!(
+                transform.transform(&parse!("(sum 10)")),
+                Ok(parse!("(+ 10)"))
+            );
+            assert_eq!(
+                transform.transform(&parse!("(sum 10 20)")),
+                Ok(parse!("(+ 10 20)"))
+            );
+            assert_eq!(
+                transform.transform(&parse!("(sum 10 20 30)")),
+                Ok(parse!("(+ 10 20 30)"))
+            );
+        }
+        {
+            let transform = Transform::try_new(&parse!(
+                r#"
+            (define-syntax square
+                  (syntax-rules ()
+                    [(_ a) (* a a)]
+            ))
+            "#
+            ))
+            .unwrap();
+            assert_eq!(
+                transform.transform(&parse!("(square 10)")),
+                Ok(parse!("(* 10 10)"))
+            );
+        }
+        {
+            let transform = Transform::try_new(&parse!(
+                r#"
+            (define-syntax square-of-sums
+                  (syntax-rules ()
+                    [(_ a1 a* ...)
+                     (* (+ a1 a* ...) (+ a1 a* ...))]
+            ))
+            "#
+            ))
+            .unwrap();
+            assert_eq!(
+                transform.transform(&parse!("(square-of-sums 10)")),
+                Ok(parse!("(* (+ 10) (+ 10))"))
+            );
+            assert_eq!(
+                transform.transform(&parse!("(square-of-sums 10 20)")),
+                Ok(parse!("(* (+ 10 20) (+ 10 20))"))
+            );
+            assert_eq!(
+                transform.transform(&parse!("(square-of-sums 10 20 30)")),
+                Ok(parse!("(* (+ 10 20 30) (+ 10 20 30))"))
+            );
+        }
+    }
+
+    #[test]
+    fn literal() {
+        let transform = Transform::try_new(&parse!(
+            r#"
+            (define-syntax sum
+                  (syntax-rules (add sub)
+                    [(math add a1 a2) (+ a1 a2)]
+                    [(math sub a1 a2) (- a1 a2)]                    
+            ))
+            "#
+        ))
+        .unwrap();
+        assert!(transform
+            .transform(&parse!("(math multiply 10 10)"))
+            .is_err());
+        assert_eq!(
+            transform.transform(&parse!("(math add 10 20)")),
+            Ok(parse!("(+ 10 20)"))
+        );
+        assert_eq!(
+            transform.transform(&parse!("(math sub 10 20)")),
+            Ok(parse!("(- 10 20)"))
+        );
+    }
+
+    #[test]
+    fn alternative_ellipsis() {
+        let transform = Transform::try_new(&parse!(
+            r#"
+            (define-syntax sum
+                  (syntax-rules * ()
+                    [(sum a* *) (+ a* *)]
+            ))
+            "#
+        ))
+        .unwrap();
+        assert_eq!(
+            transform.transform(&parse!("(sum 10 20)")),
+            Ok(parse!("(+ 10 20)"))
+        );
+    }
+
+    #[test]
+    fn underscore() {
+        let transform = Transform::try_new(&parse!(
+            r#"
+            (define-syntax sum
+                  (syntax-rules ()
+                    [(sum _ a1 _ a2) (+ a1 a2)]
+            ))
+            "#
+        ))
+        .unwrap();
+        assert!(transform.transform(&parse!("(sum)")).is_err());
+        assert!(transform.transform(&parse!("(sum 10)")).is_err());
+        assert!(transform.transform(&parse!("(sum 10 20)")).is_err());
+        assert!(transform.transform(&parse!("(sum 10 20 30 )")).is_err());
+        assert_eq!(
+            transform.transform(&parse!("(sum 10 20 30 40)")),
+            Ok(parse!("(+ 20 40)"))
+        );
+    }
+
+    #[test]
+    fn underscore_as_literal() {
+        let transform = Transform::try_new(&parse!(
+            r#"
+            (define-syntax sum
+                  (syntax-rules (_)
+                    [(sum _ a1 _ a2) (+ a1 a2)]
+            ))
+            "#
+        ))
+        .unwrap();
+        assert!(transform.transform(&parse!("(sum)")).is_err());
+        assert!(transform.transform(&parse!("(sum 10)")).is_err());
+        assert!(transform.transform(&parse!("(sum 10 20)")).is_err());
+        assert!(transform.transform(&parse!("(sum 10 20 30 )")).is_err());
+        assert!(transform.transform(&parse!("(sum 10 20 30 40)")).is_err());
+        assert_eq!(
+            transform.transform(&parse!("(sum _ 20 _ 40)")),
+            Ok(parse!("(+ 20 40)"))
+        );
+    }
+
     #[test]
     fn zip_multi() {
         let transform = Transform::try_new(&parse!(
