@@ -1,11 +1,17 @@
+use crate::number::Number;
+use crate::vm::char::write_escaped_char;
+use crate::vm::continuation::Continuation;
 use crate::vm::environment::LexicalEnvironment;
 use crate::vm::lambda::Lambda;
 use crate::vm::opcode::OpCode;
 use crate::vm::transform::Transform;
+use crate::vm::vector::Vector;
 use crate::vm::Error::ExpectedType;
 use crate::vm::{Error, Vm};
+use std::cell::RefCell;
 use std::fmt;
 use std::fmt::{Debug, Formatter};
+use std::ops::Deref;
 use std::rc::Rc;
 
 /// VCell
@@ -20,31 +26,44 @@ use std::rc::Rc;
 /// * Other runtime sentinel values, such as Undefined and Void.
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub enum VCell {
-    Acc,
-    BasePointer(usize),
-    BasePointerOffset(i64),
+    // Scheme primitive types
     Bool(bool),
-    BuiltInProc(BuiltInProc),
+    Char(char),
+    Nil,
+    Number(Number),
+    Pair(usize, usize),
+    Symbol(Rc<String>),
+    String(Rc<RefCell<String>>),
+    Vector(Rc<Vector>),
+
+    // other scheme values
+    Undefined,
+    Void,
+
+    // lambda, closure and lexical environments
+    Continuation(Rc<Continuation>),
     Closure(usize, usize),
-    GlobalEnvSlot(usize),
+    Lambda(Rc<Lambda>),
     LexicalEnv(Rc<LexicalEnvironment>),
     LexicalEnvSlot(usize),
     LexicalEnvPtr(usize, usize),
-    FixedNum(i64),
-    InstructionPointer(usize, usize),
     Macro(Rc<Transform>),
-    Lambda(Rc<Lambda>),
-    Nil,
+
+    // VM registers and stack values
+    Acc,
+    ArgumentCount(usize),
+    BasePointer(usize),
+    BasePointerOffset(i64),
+    BuiltInProc(BuiltInProc),
+    EnvironmentPointer(usize),
+    GlobalEnvSlot(usize),
+    InstructionPointer(usize, usize),
     OpCode(OpCode),
-    Pair(usize, usize),
     Ptr(usize),
-    Symbol(Rc<String>),
-    Undefined,
-    Void,
 }
 
 #[derive(Clone)]
-pub struct BuiltInProc(pub fn(&mut Vm) -> Result<(), Error>);
+pub struct BuiltInProc(pub fn(&mut Vm) -> Result<VCell, Error>);
 
 impl Debug for BuiltInProc {
     fn fmt(&self, f: &mut Formatter) -> Result<(), fmt::Error> {
@@ -61,25 +80,32 @@ impl PartialEq<Self> for BuiltInProc {
 impl Eq for BuiltInProc {}
 
 pub const ACC_TYPE_TEXT: &str = "#<acc>";
+pub const ARGUMENT_COUNT_TYPE_TEXT: &str = "#<argument-count>";
 pub const BASE_POINTER_TYPE_TEXT: &str = "#<base-pointer>";
 pub const BASE_POINTER_OFFSET_TYPE_TEXT: &str = "#<base-pointer-offset>";
 pub const BOOL_TYPE_TEXT: &str = "#<bool>";
+pub const CHAR_TYPE_TEXT: &str = "#<char>";
 pub const CLOSURE_TYPE_TEXT: &str = "#<closure>";
+pub const CONTINUATION_TYPE_TEXT: &str = "#<continuation>";
 pub const GLOBAL_ENV_SLOT_TYPE_TEXT: &str = "#<global-environment-slot>";
+pub const ENVIRONMENT_POINTER_TYPE_TEXT: &str = "#<environment-pointer>";
 pub const MACRO_TYPE_TEXT: &str = "#<macro>";
 pub const LEXICAL_ENV_TYPE_TEXT: &str = "#<lexical-environment>";
 pub const LEXICAL_ENV_TYPE_SLOT: &str = "#<lexical-environment-slot>";
 pub const LEXICAL_ENV_POINTER_TYPE_TEXT: &str = "#<lexical-environment-pointer>";
 pub const FIXEDNUM_TYPE_TEXT: &str = "#<fixednum>";
 pub const INSTRUCTION_POINTER_TYPE_TEXT: &str = "#<instruction-pointer>";
+pub const NUMBER_TYPE_TEXT: &str = "#<number>";
 pub const LAMBDA_TYPE_TEXT: &str = "#<lambda>";
 pub const NIL_TYPE_TEXT: &str = "#<nil>";
 pub const OPCODE_TYPE_TEXT: &str = "#<opcode>";
 pub const PAIR_TYPE_TEXT: &str = "#<pair>";
 pub const PTR_TYPE_TEXT: &str = "#<ptr>";
+pub const STRING_TYPE_TEXT: &str = "#<string>";
 pub const SYMBOL_TYPE_TEXT: &str = "#<symbol>";
 pub const SYSCALL_TYPE_TEXT: &str = "#<syscall>";
 pub const UNDEFINED_TYPE_TEXT: &str = "#<undefined>";
+pub const VECTOR_TYPE_TEXT: &str = "#<vector>";
 pub const VOID_TYPE_TEXT: &str = "#<void>";
 
 impl VCell {
@@ -90,30 +116,40 @@ impl VCell {
     pub fn type_text(&self) -> &'static str {
         match self {
             VCell::Acc => ACC_TYPE_TEXT,
+            VCell::ArgumentCount(_) => ARGUMENT_COUNT_TYPE_TEXT,
             VCell::BasePointer(_) => BASE_POINTER_TYPE_TEXT,
             VCell::BasePointerOffset(_) => BASE_POINTER_OFFSET_TYPE_TEXT,
             VCell::Bool(_) => BOOL_TYPE_TEXT,
+            VCell::Char(_) => CHAR_TYPE_TEXT,
+            VCell::Continuation(_) => CONTINUATION_TYPE_TEXT,
             VCell::Closure(_, _) => CLOSURE_TYPE_TEXT,
+            VCell::EnvironmentPointer(_) => ENVIRONMENT_POINTER_TYPE_TEXT,
             VCell::GlobalEnvSlot(_) => GLOBAL_ENV_SLOT_TYPE_TEXT,
             VCell::LexicalEnv(_) => LEXICAL_ENV_TYPE_TEXT,
             VCell::LexicalEnvSlot(_) => LEXICAL_ENV_TYPE_SLOT,
             VCell::LexicalEnvPtr(_, _) => LEXICAL_ENV_POINTER_TYPE_TEXT,
-            VCell::FixedNum(_) => FIXEDNUM_TYPE_TEXT,
             VCell::InstructionPointer(_, _) => INSTRUCTION_POINTER_TYPE_TEXT,
             VCell::Lambda(_) => LAMBDA_TYPE_TEXT,
             VCell::Nil => NIL_TYPE_TEXT,
+            VCell::Number(_) => NUMBER_TYPE_TEXT,
             VCell::OpCode(_) => OPCODE_TYPE_TEXT,
             VCell::Pair(_, _) => PAIR_TYPE_TEXT,
             VCell::Ptr(_) => PTR_TYPE_TEXT,
+            VCell::String(_) => STRING_TYPE_TEXT,
             VCell::Symbol(_) => SYMBOL_TYPE_TEXT,
             VCell::BuiltInProc(_) => SYSCALL_TYPE_TEXT,
             VCell::Macro(_) => MACRO_TYPE_TEXT,
             VCell::Undefined => UNDEFINED_TYPE_TEXT,
+            VCell::Vector(_) => VECTOR_TYPE_TEXT,
             VCell::Void => VOID_TYPE_TEXT,
         }
     }
 
-    pub fn syscall(func: fn(&mut Vm) -> Result<(), Error>) -> VCell {
+    pub fn number<T: Into<Number>>(num: T) -> VCell {
+        VCell::Number(num.into())
+    }
+
+    pub fn syscall(func: fn(&mut Vm) -> Result<VCell, Error>) -> VCell {
         VCell::BuiltInProc(BuiltInProc(func))
     }
 
@@ -125,10 +161,6 @@ impl VCell {
         VCell::Void
     }
 
-    pub fn fixed_num<T: Into<i64>>(val: T) -> VCell {
-        VCell::FixedNum(val.into())
-    }
-
     pub fn ptr(val: usize) -> VCell {
         VCell::Ptr(val)
     }
@@ -141,8 +173,16 @@ impl VCell {
         VCell::GlobalEnvSlot(slot.into())
     }
 
-    pub fn symbol<T: Into<String>>(ptr: T) -> VCell {
-        VCell::Symbol(Rc::new(ptr.into()))
+    pub fn string<T: Into<String>>(s: T) -> VCell {
+        VCell::String(Rc::new(RefCell::new(s.into())))
+    }
+
+    pub fn symbol<T: Into<String>>(sym: T) -> VCell {
+        VCell::Symbol(Rc::new(sym.into()))
+    }
+
+    pub fn vector<T: Into<Vec<VCell>>>(vector: T) -> VCell {
+        VCell::Vector(Rc::new(Vector::new(vector.into())))
     }
 
     pub fn lambda<T: Into<Lambda>>(lambda: T) -> VCell {
@@ -157,12 +197,16 @@ impl VCell {
         matches!(self, VCell::Bool(_))
     }
 
-    pub fn is_fixednum(&self) -> bool {
-        matches!(self, VCell::FixedNum(_))
+    pub fn is_number(&self) -> bool {
+        matches!(self, VCell::Number(_))
     }
 
-    pub fn is_number(&self) -> bool {
-        self.is_fixednum()
+    pub fn is_string(&self) -> bool {
+        matches!(self, VCell::String(_))
+    }
+
+    pub fn is_char(&self) -> bool {
+        matches!(self, VCell::Char(_))
     }
 
     pub fn is_pair(&self) -> bool {
@@ -221,6 +265,10 @@ impl VCell {
         matches!(self, VCell::Macro(_))
     }
 
+    pub fn is_vector(&self) -> bool {
+        matches!(self, VCell::Vector(_))
+    }
+
     pub fn as_opcode(&self) -> Result<OpCode, Error> {
         match self {
             VCell::OpCode(op) => Ok(op.clone()),
@@ -242,13 +290,6 @@ impl VCell {
         }
     }
 
-    pub fn as_fixed_num(&self) -> Result<i64, Error> {
-        match self {
-            VCell::FixedNum(val) => Ok(*val),
-            _ => Err(ExpectedType(FIXEDNUM_TYPE_TEXT, self.type_text())),
-        }
-    }
-
     pub fn as_symbol(&self) -> Result<&str, Error> {
         match self {
             VCell::Symbol(s) => Ok(&*s),
@@ -256,6 +297,13 @@ impl VCell {
         }
     }
 
+    pub fn as_string(&self) -> Result<&RefCell<String>, Error> {
+        match self {
+            VCell::String(s) => Ok(&*s),
+            _ => Err(ExpectedType(SYMBOL_TYPE_TEXT, self.type_text())),
+        }
+    }
+
     pub fn as_lambda(&self) -> Result<&Lambda, Error> {
         match self {
             VCell::Lambda(lambda) => Ok(&*lambda),
@@ -291,6 +339,27 @@ impl VCell {
         }
     }
 
+    pub fn as_vector(&self) -> Result<&Vector, Error> {
+        match self {
+            VCell::Vector(vector) => Ok(&*vector),
+            _ => Err(ExpectedType(VECTOR_TYPE_TEXT, self.type_text())),
+        }
+    }
+
+    pub fn as_argc(&self) -> Result<usize, Error> {
+        match self {
+            VCell::ArgumentCount(bp) => Ok(*bp),
+            _ => Err(ExpectedType(ARGUMENT_COUNT_TYPE_TEXT, self.type_text())),
+        }
+    }
+
+    pub fn as_number(&self) -> Result<&Number, Error> {
+        match self {
+            VCell::Number(num) => Ok(num),
+            _ => Err(ExpectedType(NUMBER_TYPE_TEXT, self.type_text())),
+        }
+    }
+
     pub fn as_ip(&self) -> Result<(usize, usize), Error> {
         match self {
             VCell::InstructionPointer(lambda, ip) => Ok((*lambda, *ip)),
@@ -308,6 +377,16 @@ impl VCell {
         }
     }
 
+    pub fn as_ep(&self) -> Result<usize, Error> {
+        match self {
+            VCell::EnvironmentPointer(bp) => Ok(*bp),
+            _ => Err(ExpectedType(
+                ENVIRONMENT_POINTER_TYPE_TEXT,
+                self.type_text(),
+            )),
+        }
+    }
+
     pub fn as_bp_offset(&self) -> Result<i64, Error> {
         match self {
             VCell::BasePointerOffset(offset) => Ok(*offset),
@@ -337,9 +416,27 @@ impl From<bool> for VCell {
     }
 }
 
+impl From<char> for VCell {
+    fn from(val: char) -> Self {
+        VCell::Char(val)
+    }
+}
+
 impl From<i64> for VCell {
-    fn from(val: i64) -> Self {
-        VCell::FixedNum(val)
+    fn from(num: i64) -> Self {
+        VCell::Number(Number::from(num))
+    }
+}
+
+impl From<i32> for VCell {
+    fn from(num: i32) -> Self {
+        VCell::Number(Number::from(num))
+    }
+}
+
+impl From<Number> for VCell {
+    fn from(num: Number) -> Self {
+        VCell::Number(num)
     }
 }
 
@@ -347,13 +444,16 @@ impl fmt::Display for VCell {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
             VCell::Acc => write!(f, "%acc"),
+            VCell::ArgumentCount(argc) => write!(f, "%argc[{}]", argc),
             VCell::BasePointer(bp) => write!(f, "%bp[${:02x}]", bp),
             VCell::BasePointerOffset(offset) => write!(f, "%bp[{:+}]", *offset),
             VCell::Bool(true) => write!(f, "#t"),
             VCell::Bool(false) => write!(f, "#f"),
+            VCell::Char(c) => write_escaped_char(*c, f),
             VCell::Closure(_, _) => write!(f, "#<closure>"),
+            VCell::Continuation(_) => write!(f, "#<continuation>"),
+            VCell::EnvironmentPointer(ep) => write!(f, "%ep[${:02x}]", ep),
             VCell::GlobalEnvSlot(slot) => write!(f, "genv[${:02x}]", slot),
-            VCell::FixedNum(val) => write!(f, "{}", val),
             VCell::InstructionPointer(lambda, ip) => {
                 write!(f, "%ip[${:02x}][${:02x}]", *lambda, *ip)
             }
@@ -363,12 +463,15 @@ impl fmt::Display for VCell {
             VCell::LexicalEnvPtr(env, slot) => write!(f, "env[${:02x}][${:02x}]", env, slot),
             VCell::Lambda(_) => write!(f, "#<lambda>"),
             VCell::Nil => write!(f, "()"),
+            VCell::Number(number) => write!(f, "{:?}", number),
             VCell::OpCode(val) => write!(f, "{:?}", val),
             VCell::Pair(car, cdr) => write!(f, "(${:02x} . ${:02x})", car, cdr),
             VCell::Ptr(ptr) => write!(f, "${:02x}", ptr),
+            VCell::String(s) => write!(f, "\"{}\"", (**s).borrow().deref()),
             VCell::Symbol(s) => write!(f, "{}", *s),
             VCell::BuiltInProc(_) => write!(f, "#<syscall>"),
             VCell::Undefined => write!(f, "undefined"),
+            VCell::Vector(_) => write!(f, "#<vector>"),
             VCell::Void => write!(f, "#<void>"),
         }
     }
